@@ -6,9 +6,11 @@
 # the Free Software Foundation; either version 2 of the License, or
 # (at your option) any later version.
 
+import gui.pygtkcompat
 from gettext import gettext as _
 import gtk, os
 gdk = gtk.gdk
+from glob import glob
 import pixbuflist
 import windowing
 from lib import tiledsurface, helpers
@@ -19,14 +21,13 @@ RESPONSE_SAVE_AS_DEFAULT = 1
 class Window(windowing.Dialog):
     def __init__(self, app):
         flags = gtk.DIALOG_DESTROY_WITH_PARENT
-        buttons = (_('save as default'), RESPONSE_SAVE_AS_DEFAULT,
+        buttons = (_('Save as Default'), RESPONSE_SAVE_AS_DEFAULT,
                    gtk.STOCK_OK, gtk.RESPONSE_ACCEPT)
         windowing.Dialog.__init__(self, app=app, title=_('Background'),
                                   parent=app.drawWindow, flags=flags,
                                   buttons=buttons)
 
         #set up window
-        self.connect('delete-event', self.app.hide_window_cb)
         self.connect('response', self.on_response)
 
         notebook = self.nb = gtk.Notebook()
@@ -48,33 +49,35 @@ class Window(windowing.Dialog):
         self.cs.connect('color-changed', self.color_changed_cb)
         color_vbox.pack_start(self.cs, expand=True)
 
-        b = gtk.Button(_('add color to patterns'))
+        b = gtk.Button(_('Add color to Patterns'))
         b.connect('clicked', self.add_color_to_patterns_cb)
         color_vbox.pack_start(b, expand=False)
 
     def on_response(self, dialog, response, *args):
         if response == RESPONSE_SAVE_AS_DEFAULT:
-             self.save_as_default_cb()
+            self.save_as_default_cb()
         elif response == gtk.RESPONSE_ACCEPT:
             self.hide()
 
     def color_changed_cb(self, widget):
         rgb = self.cs.get_current_color()
         rgb = rgb.red, rgb.green, rgb.blue
-        rgb = [int(x / 65535.0 * 255.0) for x in rgb] 
-        pixbuf = gdk.Pixbuf(gdk.COLORSPACE_RGB, False, 8, N, N)
+        rgb = [int(x / 65535.0 * 255.0) for x in rgb]
+        pixbuf = gui.pygtkcompat.gdk.pixbuf.new(gdk.COLORSPACE_RGB, False,
+                                                8, N, N)
         arr = helpers.gdkpixbuf2numpy(pixbuf)
         arr[:,:] = rgb
         self.set_background(pixbuf)
 
     def save_as_default_cb(self):
         pixbuf = self.current_background_pixbuf
-        pixbuf.save(os.path.join(self.app.confpath, 'backgrounds', 'default.png'), 'png')
+        path = os.path.join(self.app.confpath, 'backgrounds', 'default.png')
+        gui.pygtkcompat.gdk.pixbuf.save(pixbuf, path, 'png')
         self.hide()
 
     def set_background(self, pixbuf):
         doc = self.app.doc.model
-        doc.set_background(pixbuf)
+        doc.set_background(pixbuf, make_default=True)
         self.current_background_pixbuf = pixbuf
 
     def add_color_to_patterns_cb(self, widget):
@@ -103,7 +106,7 @@ class BackgroundList(pixbuflist.PixbufList):
         self.backgrounds = []
 
         def listdir(path):
-            l = [os.path.join(path, filename) for filename in os.listdir(path)]
+            l = glob(os.path.join(path, '*.png')) + glob(os.path.join(path, '*/*.png'))
             l.sort(key=os.path.getmtime)
             return l
 
@@ -111,20 +114,33 @@ class BackgroundList(pixbuflist.PixbufList):
         files.sort()
         files += listdir(user_path)
 
+        load_errors = []
         for filename in files:
             if not filename.lower().endswith('.png'):
                 continue
-            pixbuf = gdk.pixbuf_new_from_file(filename)
-
-            # error checking
-            def error(msg):
-                self.app.message_dialog(msg, type = gtk.MESSAGE_WARNING, flags = gtk.DIALOG_MODAL)
-            if pixbuf.get_has_alpha():
-                error(_('The background %s was ignored because it has an alpha channel. Please remove it.') % filename)
+            try:
+                pixbuf = gdk.pixbuf_new_from_file(filename)
+            except Exception, ex:
+                print ex
+                load_errors.append(
+                    _('Gdk-Pixbuf couldn\'t load "{filename}", and reported "{error}"').format(
+                    filename=filename, error=repr(ex)))
                 continue
+            supported = True
+            if pixbuf.get_has_alpha():
+                load_errors.append(
+                    _('"%s" has an alpha channel. Background images with '
+                      'transparency are not supported.')
+                    % filename)
+                supported = False
             w, h = pixbuf.get_width(), pixbuf.get_height()
             if w % N != 0 or h % N != 0 or w == 0 or h == 0:
-                error(_('The background %s was ignored because it has the wrong size. Only (N*%d)x(M*%d) is supported.') % (filename, N, N))
+                load_errors.append(
+                    _('{filename} has an unsupported size. Background images '
+                      'must have widths and heights which are multiples '
+                      'of {number} pixels.').format(filename=filename, number=N))
+                supported = False
+            if not supported:
                 continue
 
             if os.path.basename(filename).lower() == 'default.png':
@@ -133,6 +149,17 @@ class BackgroundList(pixbuflist.PixbufList):
 
             self.backgrounds.append(pixbuf)
 
+        if load_errors:
+            msg = "\n\n".join(load_errors)
+            self.app.message_dialog(
+                text=_("One or more backgrounds could not be loaded"),
+                title=_("Error loading backgrounds"),
+                secondary_text=_("Please remove the unloadable files, or "
+                                 "check your libgdkpixbuf installation."),
+                long_text=msg,
+                type=gtk.MESSAGE_WARNING,
+                flags=gtk.DIALOG_MODAL)
+
         pixbuflist.PixbufList.__init__(self, self.backgrounds, N, N, pixbuffunc=self.pixbuf_scaler)
         self.dragging_allowed = False
 
