@@ -10,6 +10,7 @@
 This module does file management for brushes and brush groups.
 """
 
+import pygtkcompat
 import dialogs
 import gtk
 from gtk import gdk # only for gdk.pixbuf
@@ -17,16 +18,15 @@ from gettext import gettext as _
 import os, zipfile
 from os.path import basename
 import urllib
-import gobject
 from lib.brush import BrushInfo
 from warnings import warn
 
 preview_w = 128
 preview_h = 128
 
-DEFAULT_STARTUP_GROUP = 'set#2'  # Suggestion only
-DEFAULT_BRUSH = 'deevad/artpen'  # TODO: phase out and use heuristics?
-DEFAULT_ERASER = 'deevad/stick'  # TODO: ---------------"--------------
+DEFAULT_STARTUP_GROUP = 'set#2'  # Suggestion only (FIXME: no effect?)
+DEFAULT_BRUSH = 'deevad/2B_pencil'  # TODO: phase out and use heuristics?
+DEFAULT_ERASER = 'deevad/kneaded_eraser_large'  # TODO: ---------------"--------------
 FOUND_BRUSHES_GROUP = 'lost&found'
 DELETED_BRUSH_GROUP = 'deleted'
 FAVORITES_BRUSH_GROUP = 'favorites'
@@ -215,9 +215,9 @@ class BrushManager:
         self.history = [None for i in xrange(BRUSH_HISTORY_SIZE)]
 
         brush_by_name = {}
-        def get_brush(name):
+        def get_brush(name, **kwargs):
             if name not in brush_by_name:
-                b = ManagedBrush(self, name, persistent=True)
+                b = ManagedBrush(self, name, persistent=True, **kwargs)
                 brush_by_name[name] = b
             return brush_by_name[name]
 
@@ -299,25 +299,22 @@ class BrushManager:
         # Distinguish between brushes in the brushlist and those that are not;
         # handle lost-and-found ones.
         for name in listbrushes(self.stock_brushpath) + listbrushes(self.user_brushpath):
-            b = get_brush(name)
-            b.in_brushlist = True
             if name.startswith('context'):
+                b = get_brush(name)
                 i = int(name[-2:])
                 self.contexts[i] = b
-                b.load_settings(retain_parent=True)
-                b.in_brushlist = False
             elif name.startswith(DEVBRUSH_NAME_PREFIX):
+                b = get_brush(name)
                 device_name = devbrush_unquote(name)
                 self.brush_by_device[device_name] = b
-                b.load_settings(retain_parent=True)
-                b.in_brushlist = False
             elif name.startswith(BRUSH_HISTORY_NAME_PREFIX):
+                b = get_brush(name)
                 i_str = name.replace(BRUSH_HISTORY_NAME_PREFIX, '')
                 i = int(i_str)
                 self.history[i] = b
-                b.load_settings(retain_parent=True)
-                b.in_brushlist = False
-            if b.in_brushlist:
+            else:
+                # normal brush that will appear in the brushlist
+                b = get_brush(name)
                 if not [True for group in our.itervalues() if b in group]:
                     brushes = self.groups.setdefault(FOUND_BRUSHES_GROUP, [])
                     brushes.insert(0, b)
@@ -477,7 +474,6 @@ class BrushManager:
                     preview_f.write(preview_data)
                     preview_f.close()
                     b.load()
-                    b.in_brushlist = True
                 # finally, add it to the group
                 if b not in managed_brushes:
                     managed_brushes.append(b)
@@ -542,16 +538,20 @@ class BrushManager:
         for i, h in enumerate(self.history):
             h.name = u"%s%d" % (BRUSH_HISTORY_NAME_PREFIX, i)
 
+    def is_in_brushlist(self, brush):
+        """Returns whether this brush is accessible through the brush selector."""
+        for group, brushes in self.groups.iteritems():
+            if brush in brushes:
+                return True
+        return False
 
     def select_brush(self, brush):
         """Selects a ManagedBrush, highlights it, & updates the live brush."""
         if brush is None:
             brush = self.get_default_brush()
-        if brush.persistent and not brush.settings_loaded:   # XXX refactor
-            brush.load_settings()
 
         brushinfo = brush.brushinfo
-        if not brush.in_brushlist:
+        if not self.is_in_brushlist(brush):
             # select parent brush instead, but keep brushinfo
             brush = self.get_parent_brush(brush=brush)
             if not brush:
@@ -569,8 +569,6 @@ class BrushManager:
         """Gets the parent `ManagedBrush` for a brush or a `BrushInfo`.
         """
         if brush is not None:
-            if brush.persistent and not brush.settings_loaded:   # XXX refactor
-                brush.load_settings()
             brushinfo = brush.brushinfo
         if brushinfo is None:
             raise RuntimeError, "One of `brush` or `brushinfo` must be defined."
@@ -581,8 +579,6 @@ class BrushManager:
             parent_brush = self.get_brush_by_name(parent_name)
             if parent_brush is None:
                 return None
-            if parent_brush.persistent and not parent_brush.settings_loaded:  # XXX refactor
-                parent_brush.load_settings()
             return parent_brush
 
 
@@ -680,10 +676,9 @@ class ManagedBrush(object):
         self.bm = brushmanager
         self._preview = None
         self.name = name
-        self.brushinfo = BrushInfo()
+        self._brushinfo = BrushInfo()
         self.persistent = persistent #: If True this brush is stored in the filesystem.
         self.settings_loaded = False  #: If True this brush is fully initialized, ready to paint with.
-        self.in_brushlist = False  #: Set to True if this brush is known to be in the brushlist
 
         self.settings_mtime = None
         self.preview_mtime = None
@@ -697,21 +692,30 @@ class ManagedBrush(object):
         if self._preview is None and self.name:
             self._load_preview()
         if self._preview is None:
-            self.preview = gdk.Pixbuf(gdk.COLORSPACE_RGB, False, 8, preview_w, preview_h)
+            # When does this happen?
+            self.preview = pygtkcompat.gdk.pixbuf.new(gdk.COLORSPACE_RGB,
+                            False, 8, preview_w, preview_h)
             self.preview.fill(0xffffffff) # white
         return self._preview
     def set_preview(self, pixbuf):
         self._preview = pixbuf
     preview = property(get_preview, set_preview)
 
+    # load brush settings on demand
+    def get_brushinfo(self):
+        if self.persistent and not self.settings_loaded:
+            self._load_settings()
+        return self._brushinfo
+    def set_brushinfo(self, brushinfo):
+        self._brushinfo = brushinfo
+    brushinfo = property(get_brushinfo, set_brushinfo)
+
     def get_display_name(self):
         """Gets a displayable name for the brush.
         """
-        if self.in_brushlist:
+        if self.bm.is_in_brushlist(self):  # FIXME: get rid of this check
             dname = self.name
         else:
-            if self.persistent and not self.settings_loaded:   # XXX refactor
-                self.load_settings()
             dname = self.brushinfo.get_string_property("parent_brush_name")
         if dname is None:
             return _("Unknown Brush")
@@ -758,7 +762,7 @@ class ManagedBrush(object):
         if not self.settings_loaded:   # XXX refactor
             self.load()
         target.brushinfo = self.brushinfo.clone()
-        if self.in_brushlist:
+        if self.bm.is_in_brushlist(self): # FIXME: get rid of this check!
             target.brushinfo.set_string_property("parent_brush_name", self.name)
         target.preview = self.preview
         target.name = name
@@ -790,19 +794,29 @@ class ManagedBrush(object):
 
     def save(self):
         prefix = self.get_fileprefix(saving=True)
-        self.preview.save(prefix + '_prev.png', 'png')
+
+        if self.preview.get_has_alpha():
+            # remove it (previous mypaint versions would display an empty image)
+            w, h = preview_w, preview_h
+            tmp = pygtkcompat.gdk.pixbuf.new(gdk.COLORSPACE_RGB, False,
+                                             8, w, h)
+            tmp.fill(0xffffffff)
+            self.preview.composite(tmp, 0, 0, w, h, 0, 0, 1, 1, gdk.INTERP_BILINEAR, 255)
+            self.preview = tmp
+
+        pygtkcompat.gdk.pixbuf.save(self.preview, prefix + '_prev.png', 'png')
         brushinfo = self.brushinfo.clone()
         open(prefix + '.myb', 'w').write(brushinfo.save_to_string())
         self.remember_mtimes()
 
-    def load(self, retain_parent=False):
+    def load(self):
         """Loads the brush's preview and settings from disk."""
         if self.name is None:
             warn("Attempt to load an unnamed brush, don't do that.",
                  RuntimeWarning, 2)
             return
         self._load_preview()
-        self.load_settings(retain_parent)
+        self._load_settings()
 
     def _load_preview(self):
         """Loads the brush preview as pixbuf into the brush."""
@@ -814,16 +828,16 @@ class ManagedBrush(object):
         self._preview = pixbuf
         self.remember_mtimes()
 
-    def load_settings(self, retain_parent=False):
+    def _load_settings(self):
         """Loads the brush settings/dynamics from disk."""
         prefix = self.get_fileprefix()
         filename = prefix + '.myb'
         brushinfo_str = open(filename).read()
-        self.brushinfo.load_from_string(brushinfo_str)
+        self._brushinfo.load_from_string(brushinfo_str)
         self.remember_mtimes()
         self.settings_loaded = True
-        if not retain_parent:
-            self.brushinfo.set_string_property("parent_brush_name", None)
+        if self.bm.is_in_brushlist(self): # FIXME: get rid of this check
+            self._brushinfo.set_string_property("parent_brush_name", None)
         self.persistent = True
 
     def reload_if_changed(self):
@@ -835,11 +849,11 @@ class ManagedBrush(object):
         self.load()
         return True
 
-    def __str__(self):
-        if self.brushinfo.settings:
-            return "<ManagedBrush %s p=%s>" % (self.name, self.brushinfo.get_string_property("parent_brush_name"))
+    def __repr__(self):
+        if self._brushinfo.settings:
+            return "<ManagedBrush %r p=%s>" % (self.name, self._brushinfo.get_string_property("parent_brush_name"))
         else:
-            return "<ManagedBrush %s (settings not loaded yet)>" % self.name
+            return "<ManagedBrush %r (settings not loaded yet)>" % self.name
 
 
 if __name__ == '__main__':
