@@ -6,22 +6,33 @@
 # the Free Software Foundation; either version 2 of the License, or
 # (at your option) any later version.
 
-"select brush window"
+"""Brush selection/organizer window
+
+Can be used for selecting brushes, and can be docked into the sidebar.
+Responsible for ordering, loading and saving brush lists.
+
+"""
+
+import pygtkcompat
+if pygtkcompat.USE_GTK3:
+    import gi
+    from gi.repository import PangoCairo
+
 import gtk, pango
 gdk = gtk.gdk
 from gettext import gettext as _
 
-import windowing
-import stock
 import pixbuflist, dialogs, brushmanager
 from elastic import ElasticExpander
 from brushlib import brushsettings
+from lib.helpers import escape
 
 class ToolWidget (gtk.VBox):
 
     EXPANDER_PREFS_KEY = "brushmanager.common_settings_expanded"
 
-    stock_id = stock.TOOL_BRUSH
+    stock_id = "mypaint-tool-brush"
+    tool_widget_title = _("Brushes")
 
     def __init__(self, app):
         self.app = app
@@ -52,7 +63,7 @@ class ToolWidget (gtk.VBox):
         self.expander.set_size_request(48, -1)
 
     def show_cb(self, widget):
-        assert not self.window
+        assert self.get_window() is None
         assert not self.expander_prefs_loaded
         is_expanded = bool(self.app.preferences.get(self.EXPANDER_PREFS_KEY, False))
         self.expander.set_expanded(is_expanded)
@@ -108,7 +119,9 @@ class BrushList(pixbuflist.PixbufList):
                                        pixbuffunc = lambda x: x.preview)
         # Support device changing with the same event as that used
         # for brush choice:
-        self.set_extension_events(gdk.EXTENSION_EVENTS_ALL)
+        if not pygtkcompat.USE_GTK3:
+            self.set_extension_events(gdk.EXTENSION_EVENTS_ALL)
+
         self.set_selected(self.bm.selected_brush)
         self.bm.brushes_observers.append(self.brushes_modified_cb)
         self.bm.selected_brush_observers.append(self.brush_selected_cb)
@@ -129,8 +142,8 @@ class BrushList(pixbuflist.PixbufList):
         for f in self.bm.brushes_observers: f(self.brushes)
 
     def button_press_cb(self, widget, event):
-        self.app.doc.tdw.device_used(event.device)
-        pixbuflist.PixbufList.button_press_cb(self, widget, event)
+        self.app.device_monitor.device_used(event.device)
+        return pixbuflist.PixbufList.button_press_cb(self, widget, event)
 
     def drag_begin_cb(self, widget, context):
         preview = self.bm.selected_brush.preview
@@ -184,7 +197,8 @@ class BrushGroupsList(gtk.VBox):
                 # (no real problem, this is only when deleting/renaming groups)
                 del self.group_widgets[group]
 
-        self.foreach(self.remove)
+        for child in list(self.get_children()):
+            self.remove(child)
 
         for group in self.bm.active_groups:
             if group in self.group_widgets:
@@ -207,16 +221,22 @@ class GroupSelector(gtk.DrawingArea):
         self.bm = app.brushmanager
         self.bm.groups_observers.append(self.active_groups_changed_cb)
 
-        self.drag_dest_set(gtk.DEST_DEFAULT_MOTION | gtk.DEST_DEFAULT_DROP,
+        if not pygtkcompat.USE_GTK3:
+            self.drag_dest_set(gtk.DEST_DEFAULT_MOTION | gtk.DEST_DEFAULT_DROP,
                 [('LIST_ITEM', gtk.TARGET_SAME_APP, pixbuflist.DRAG_ITEM_NAME)],
                 gdk.ACTION_COPY|gdk.ACTION_MOVE)
+
         self.connect('drag-motion', self.drag_motion_cb)
         self.connect('drag-data-received', self.drag_data_received_cb)
         self.connect('drag-leave', self.drag_clear_cb)
         self.connect('drag-begin', self.drag_clear_cb)
         self.connect('drag-end', self.drag_clear_cb)
 
-        self.connect("expose-event", self.expose_cb)
+        if pygtkcompat.USE_GTK3:
+            self.connect("draw", self.draw_cb)
+        else:
+            self.connect("expose-event", self.expose_cb)
+
         self.connect("button-press-event", self.button_press_cb)
         self.connect("motion-notify-event", self.motion_notify_cb)
         self.connect("leave-notify-event", self.leave_notify_cb)
@@ -230,8 +250,13 @@ class GroupSelector(gtk.DrawingArea):
         self.gtkstate_prelight_group = None
         self.gtkstate_active_group = None
         self.drag_target_group = None
-        self.set_tooltip_text(_('Try right click, middle click or Ctrl click'))
-        self.connect("size-request", self.on_size_request)
+        self.set_tooltip_text(_('Brush Groups: click to select\n'
+                                'Ctrl: select multiple\n'
+                                'Middle-click: toggle group\n'
+                                'Right-click: groups menu'))
+
+        if not pygtkcompat.USE_GTK3:
+            self.connect("size-request", self.on_size_request)
 
     def active_groups_changed_cb(self):
         self.queue_draw()
@@ -246,7 +271,6 @@ class GroupSelector(gtk.DrawingArea):
 
         all_groups = list(sorted(self.bm.groups.keys()))
         idx = 0
-        text = ''
         attr = pango.AttrList()
         self.idx2group = {}
 
@@ -256,17 +280,22 @@ class GroupSelector(gtk.DrawingArea):
         sp_s = pad_s + u"\u200b"  # ZERO WIDTH SPACE
 
         import platform
-        if platform.system() == 'Windows':
+        if platform.system() == 'Windows' or platform.system() == 'Darwin':
             # workaround for https://gna.org/bugs/?15192
             pad_s = ''
             sp_s = ' '
 
+        def _gdk_color_to_hex(col):
+            rgb = (col.red, col.green, col.blue)
+            rgb = [max(min(c>>8, 255), 0) for c in rgb]
+            return "#%02x%02x%02x" % tuple(rgb)
+
+        markup = ''
         for group in all_groups:
             group_label = brushmanager.translate_group_name(group)
             u = pad_s + group_label + pad_s
-            s = u.encode('utf-8')
             idx_start = idx
-            for c in s:
+            for c in u.encode('utf-8'):
                 self.idx2group[idx] = group
                 idx += 1
 
@@ -281,42 +310,62 @@ class GroupSelector(gtk.DrawingArea):
 
             style_fg, style_bg = style.fg, style.bg
             if group == self.drag_target_group:
-                # Invert colurs
+                # Invert colours
                 style_fg, style_bg = style.bg, style.fg
-                # attr.insert(pango.AttrUnderline(pango.UNDERLINE_SINGLE, idx_start, idx))
 
             # always use the STATE_SELECTED fg if the group is visible
             if group in self.bm.active_groups:
                 fg_state = gtk.STATE_SELECTED
 
-            c = style_bg[bg_state]
-            attr.insert(pango.AttrBackground(c.red, c.green, c.blue, idx_start, idx))
-            c = style_fg[fg_state]
-            attr.insert(pango.AttrForeground(c.red, c.green, c.blue, idx_start, idx))
-
-            text += u + sp_s
+            c_bg = _gdk_color_to_hex(style_bg[bg_state])
+            c_fg = _gdk_color_to_hex(style_fg[fg_state])
+            m = escape(u)
+            m = "<span fgcolor='%s' bgcolor='%s'>%s</span>" % (c_fg, c_bg, m)
+            markup += m + sp_s
             idx += len(sp_s.encode("utf-8"))
 
-        layout.set_text(text)
-        layout.set_attributes(attr)
-
+        layout.set_markup(markup)
         leading = style.font_desc.get_size() / 6
         layout.set_spacing(leading)
         return layout
 
     def on_size_request(self, widget, req):
+        # PyGTK/gtk2 size-request handler
         parent = self.parent.parent
         parent_width = parent.get_allocation().width
-        # The above is potentially the adopt "natural size" at first, but
-        # we should respect it if it's set. Might result in fewer redraws.
+        # The above is potentially the "adopt natural size" value, -1, at
+        # first, but we should respect it if it's set. Doing so might result
+        # in fewer redraws.
         layout = self.lay_out_group_names(parent_width)
         w, h = layout.get_pixel_size()
         h += 2 * self.VERTICAL_MARGIN
         req.width = -1
         req.height = h
 
+    def do_get_request_mode(self):
+        # PyGI/gtk3 sizing
+        # Width first, then height
+        return gtk.SizeRequestMode.HEIGHT_FOR_WIDTH
+
+    def do_get_preferred_width(self):
+        # PyGI/gtk3 sizing
+        # Enough width to lay out 1 or 2 short words, but ideally a little
+        # more. In most cases, the parent determines everything.
+        return (50, 100)
+
+    def do_get_preferred_height_for_width(self, width):
+        # PyGI/gtk3 sizing
+        # Height is determined by the layout once we have a concrrete width.
+        layout = self.lay_out_group_names(width)
+        w, h = layout.get_pixel_size()
+        h += 2 * self.VERTICAL_MARGIN
+        return (h, h)
+
     def expose_cb(self, widget, event):
-        cr = self.window.cairo_create()
+        cr = self.get_window().cairo_create()
+        return self.draw_cb(widget, cr)
+
+    def draw_cb(self, widget, cr):
         alloc = self.get_allocation()
         width = alloc.width
         height = alloc.height
@@ -324,12 +373,14 @@ class GroupSelector(gtk.DrawingArea):
         style = self.get_style()
 
         c = style.bg[gtk.STATE_NORMAL]
-        cr.set_source_rgb(c.red_float, c.green_float, c.blue_float)
+        c_rgb = [float(x)/65535 for x in (c.red, c.green, c.blue)]
+        cr.set_source_rgb(*c_rgb)
         cr.rectangle(0, 0, width, height)
         cr.fill()
 
         c = style.text[gtk.STATE_NORMAL]
-        cr.set_source_rgb(c.red_float, c.green_float, c.blue_float)
+        c = [float(x)/65535 for x in (c.red, c.green, c.blue)]
+        cr.set_source_rgb(*c_rgb)
         layout = self.lay_out_group_names(width)
 
         leading = style.font_desc.get_size() / 6
@@ -337,7 +388,10 @@ class GroupSelector(gtk.DrawingArea):
         layout.set_spacing(leading)
 
         cr.move_to(0, self.VERTICAL_MARGIN)
-        cr.show_layout(layout)
+        if pygtkcompat.USE_GTK3:
+            PangoCairo.show_layout(cr, layout)
+        else:
+            cr.show_layout(layout)
 
         # Catch reflows, which maybe result in more or fewer rows.
         self.set_size_request(-1, -1)  # "ask again, give me my natural size"
@@ -349,7 +403,11 @@ class GroupSelector(gtk.DrawingArea):
         x, y = int(x), int(y) # avoid warning
         if self.layout is None:
             return None
-        i, d = self.layout.xy_to_index(x*pango.SCALE, y*pango.SCALE)
+        index_tup = self.layout.xy_to_index(x*pango.SCALE, y*pango.SCALE)
+        if pygtkcompat.USE_GTK3:
+            inside, i, trailing = index_tup
+        else:
+            i, trailing = index_tup
         return self.idx2group.get(i)
 
     def button_press_cb(self, widget, event):
@@ -380,7 +438,13 @@ class GroupSelector(gtk.DrawingArea):
             self.gtkstate_active_group = group
             self.queue_draw()
             menu = self.context_menu(group)
-            menu.popup(None,None,None, event.button, event.time, group)
+            data = None
+            def _pos(*a):
+                return int(event.x_root), int(event.y_root), True
+            if pygtkcompat.USE_GTK3:
+                menu.popup(None, None, _pos, data, event.button, event.time)
+            else:
+                menu.popup(None, None, _pos, event.button, event.time, data)
 
     def motion_notify_cb(self, widget, event):
         old_prelight_group = self.gtkstate_prelight_group
@@ -398,18 +462,25 @@ class GroupSelector(gtk.DrawingArea):
     def context_menu(self, group):
         m = gtk.Menu()
         menu = []
-        menu += [ (_("New group..."), self.create_group_cb) ]
+        menu += [ (_("New Group..."), self.create_group_cb) ]
         if group:
-            menu += [ (_("Rename group..."), self.rename_group_cb),
-                      (_("Delete group..."), self.delete_group_cb),
-                      (_("Export group as brush package..."), self.export_group_cb),
-                      ]
-        menu += [ (_("Download brushes (in web browser)"), self.app.drawWindow.download_brush_pack_cb) ]
-        menu += [ (_("Import brush package..."), self.app.drawWindow.import_brush_pack_cb) ]
-        for label, callback in menu:
-            mi = gtk.MenuItem(label)
-            mi.connect('activate', callback, group)
-            m.append(mi)
+            menu += [ (_("Rename Group..."), self.rename_group_cb),
+                      (_("Delete Group"), self.delete_group_cb),
+                      (_("Export Group..."), self.export_group_cb), ]
+        menu += [ None ]
+        menu += [ (_("Import Brushes..."),
+                   self.app.drawWindow.import_brush_pack_cb) ]
+        menu += [ (_("Get More Brushes..."),
+                   self.app.drawWindow.download_brush_pack_cb) ]
+        for entry in menu:
+            if entry is None:
+                item = gtk.SeparatorMenuItem()
+
+            else:
+                label, callback = entry
+                item = gtk.MenuItem(label)
+                item.connect('activate', callback, group)
+            m.append(item)
         m.connect('selection-done', self.menu_finished_cb)
         m.show_all()
         return m
@@ -419,12 +490,12 @@ class GroupSelector(gtk.DrawingArea):
         self.queue_draw()
 
     def create_group_cb(self, w, group):
-        new_group = dialogs.ask_for_name(self, _('Create group'), '')
+        new_group = dialogs.ask_for_name(self, _('Create Group'), '')
         if new_group:
             self.bm.create_group(new_group)
 
     def rename_group_cb(self, w, old_group):
-        new_group = dialogs.ask_for_name(self, _('Rename group'), old_group)
+        new_group = dialogs.ask_for_name(self, _('Rename Group'), old_group)
         if not new_group:
             return
         if new_group not in self.bm.groups:
@@ -433,7 +504,7 @@ class GroupSelector(gtk.DrawingArea):
             dialogs.error(self, _('A group with this name already exists!'))
 
     def export_group_cb(self, w, group):
-        format_id, filename = dialogs.save_dialog(_("Export brush pack..."), None,
+        format_id, filename = dialogs.save_dialog(_("Export Brushes"), None,
                                  [(_("MyPaint brush package (*.zip)"), "*.zip")],
                                  default_format = (0, ".zip"))
         if filename is not None:
