@@ -16,8 +16,7 @@ from gettext import ngettext
 
 from lib import document, helpers, tiledsurface
 import drawwindow
-
-import mimetypes
+import pygtkcompat
 
 SAVE_FORMAT_ANY = 0
 SAVE_FORMAT_ORA = 1
@@ -55,25 +54,7 @@ class FileHandler(object):
         #NOTE: filehandling and drawwindow are very tightly coupled
         self.save_dialog = None
 
-        file_actions = [ \
-        ('New',          gtk.STOCK_NEW, _('New'), '<control>N', _("Start with a new blank canvas"), self.new_cb),
-        ('Open',         gtk.STOCK_OPEN, _('Open...'), '<control>O', _("Open a file, replacing the current painting"), self.open_cb),
-        ('OpenLast',     None, _('Open Last'), 'F3', _("Open the last file you saved"), self.open_last_cb),
-        ('Reload',       gtk.STOCK_REFRESH, _('Reload'), 'F5', _("Reload the current file"), self.reload_cb),
-        ('Save',         gtk.STOCK_SAVE, _('Save'), '<control>S', _("Save to a file"), self.save_cb),
-        ('SaveAs',       gtk.STOCK_SAVE_AS, _('Save As...'), '<control><shift>S', _("Save to a file with a new name"), self.save_as_cb),
-        ('Export',       gtk.STOCK_SAVE_AS, _('Export...'), '<control><shift>E', _("Export to a file"), self.save_as_cb),
-        ('SaveScrap',    gtk.STOCK_ADD, _('Save As Scrap'), 'F2',
-            _("Saves to a new scrap file. "
-              "If this drawing is currently saved as a scrap, this will "
-              "write a new revision of it."),
-            self.save_scrap_cb),
-        ('PrevScrap',    gtk.STOCK_GO_BACK, _('Open Previous Scrap'), 'F6', _("Load the previous scrap file, replacing the current drawing"), self.open_scrap_cb),
-        ('NextScrap',    gtk.STOCK_GO_FORWARD, _('Open Next Scrap'), 'F7', _("Load the next scrap file, replacing the current drawing"), self.open_scrap_cb),
-        ]
-        ag = gtk.ActionGroup('FileActions')
-        ag.add_actions(file_actions)
-        self.app.add_action_group(ag)
+        ag = app.builder.get_object('FileActions')
 
         ra = gtk.RecentAction('OpenRecent', _('Open Recent'), _('Open Recent files'), None)
         ra.set_show_tips(True)
@@ -128,7 +109,7 @@ class FileHandler(object):
         # with utf-8 characters into this list, I assume this is a
         # gtk bug.  So we use our own test instead of i.exists().
         self.recent_items = [
-                i for i in gtk.recent_manager_get_default().get_items()
+                i for i in pygtkcompat.gtk.recent_manager_get_default().get_items()
                 if "mypaint" in i.get_applications() and os.path.exists(helpers.uri2filename(i.get_uri()))
         ]
         self.recent_items.reverse()
@@ -210,7 +191,7 @@ class FileHandler(object):
         b = d.add_button(_("_Save as Scrap"), gtk.RESPONSE_APPLY)
         b.set_image(gtk.image_new_from_stock(gtk.STOCK_SAVE, gtk.ICON_SIZE_BUTTON))
 
-        d.set_has_separator(False)
+        # d.set_has_separator(False)
         d.set_default_response(gtk.RESPONSE_CANCEL)
         l = gtk.Label()
         l.set_markup("<b>%s</b>\n\n%s" % (question,t))
@@ -227,19 +208,20 @@ class FileHandler(object):
     def new_cb(self, action):
         if not self.confirm_destructive_action():
             return
-        bg = self.doc.model.background
         self.doc.model.clear()
-        self.doc.model.set_background(bg)
         # Match scratchpad to canvas background
         # TODO make this into a preference
         if self.app.scratchpad_doc:
-            self.app.scratchpad_doc.model.set_background(bg)
+            self.app.scratchpad_doc.model.set_background(self.doc.model.background)
         self.filename = None
         self.set_recent_items()
         self.app.doc.reset_view_cb(None)
 
     @staticmethod
     def gtk_main_tick():
+        if pygtkcompat.USE_GTK3:
+            # FIXME: use something better
+            return
         while gtk.events_pending():
             gtk.main_iteration(False)
 
@@ -278,26 +260,26 @@ class FileHandler(object):
         thumbnail_pixbuf = self.save_doc_to_file(filename, self.doc, export=export, **options)
         if not export:
             self.filename = os.path.abspath(filename)
-            gtk.recent_manager_get_default().add_full(helpers.filename2uri(self.filename),
-                    {
-                        'app_name': 'mypaint',
-                        'app_exec': sys.argv_unicode[0].encode('utf-8'),
-                        # todo: get mime_type
-                        'mime_type': 'application/octet-stream'
-                    }
-            )
-
+            recent_mgr = pygtkcompat.gtk.recent_manager_get_default()
+            uri = helpers.filename2uri(self.filename)
+            recent_data = dict(app_name='mypaint',
+                               app_exec=sys.argv_unicode[0].encode('utf-8'),
+                               # todo: get mime_type
+                               mime_type='application/octet-stream')
+            if pygtkcompat.USE_GTK3:
+                # No Gtk.RecentData.new() as of 3.4.2-0ubuntu0.3,
+                # nor can we set the fields of an empty one :(
+                recent_mgr.add_item(uri)
+            else:
+                recent_mgr.add_full(uri, recent_data)
         if not thumbnail_pixbuf:
             thumbnail_pixbuf = self.doc.model.render_thumbnail()
-        helpers.freedesktop_thumbnail(self.filename, thumbnail_pixbuf)
+        helpers.freedesktop_thumbnail(filename, thumbnail_pixbuf)
 
     @drawwindow.with_wait_cursor
     def save_scratchpad(self, filename, export=False, **options):
-        # Check that there is something to save:
-        x, y, w, h =  self.app.scratchpad_doc.model.get_bbox()
-        if w == 0 and h == 0:
-            w, h = tiledsurface.N, tiledsurface.N # TODO: support for other sizes
-        thumbnail_pixbuf = self.save_doc_to_file(filename, self.app.scratchpad_doc, export=export, **options)
+        if self.app.scratchpad_doc.model.unsaved_painting_time or export or not os.path.exists(filename):
+            self.save_doc_to_file(filename, self.app.scratchpad_doc, export=export, **options)
         if not export:
             self.app.scratchpad_filename = os.path.abspath(filename)
             self.app.preferences["scratchpad.last_opened_scratchpad"] = self.app.scratchpad_filename
@@ -490,7 +472,7 @@ class FileHandler(object):
                         except KeyError:
                             saveformat = default_saveformat
                     else:
-                            saveformat = default_saveformat
+                        saveformat = default_saveformat
 
                 desc, ext_format, options = self.saveformats[saveformat]
 
