@@ -11,17 +11,29 @@ Tools can be snapped in and out of the sidebar.
 Window sizes and sidebar positions are stored to user preferences.
 """
 
+import os
+
+import pygtkcompat
 import gtk
-import gobject
 from gtk import gdk
-from math import sqrt
-from warnings import warn
 import pango
 import cairo
+gobject = pygtkcompat.gobject
+
+from math import sqrt
+from warnings import warn
+
 from gettext import gettext as _
 
 from elastic import ElasticContainer
 
+def style_get_property(widget, property_name):
+
+    if pygtkcompat.USE_GTK3:
+        # FIXME: is the equivalent GI call broken? It seems to take value as an argument..
+        return 10
+    else:
+        return widget.style_get_property(property_name)
 
 class LayoutManager:
     """Keeps track of tool positions, and main window state.
@@ -82,7 +94,10 @@ class LayoutManager:
         if role in self.widgets:
             return self.widgets[role]
         else:
-            result = self.factory(role, self, *self.factory_opts)
+            try:
+                result = self.factory(role, self, *self.factory_opts)
+            except ImportError:
+                result = None
             if role == 'main-window':
                 self.main_window = result[0]
                 assert isinstance(self.main_window, MainWindow)
@@ -108,9 +123,7 @@ class LayoutManager:
                     widget.connect("hide", cb, False)
                     return widget
                 elif isinstance(widget, gtk.Widget):
-                    stock_id = result[1]
-                    stock_info = gtk.stock_lookup(stock_id)
-                    stock_id, title, _mods, _key, _tr = stock_info
+                    stock_id, title = result[1], result[2]
                     tool = Tool(widget, role, title, stock_id, self)
                     self.tools[role] = tool
                     self.widgets[role] = tool
@@ -301,7 +314,10 @@ class WindowWithSavedPosition:
         """Store the current size of the window for future launches.
         """
         # Save the new position in the prefs...
-        f_ex = self.window.get_frame_extents()
+        if not hasattr(self, 'get_window'):
+            return
+
+        f_ex = self.get_window().get_frame_extents()
         x = max(0, f_ex.x)
         y = max(0, f_ex.y)
         conf_pos = dict(x=x, y=y, w=event.width, h=event.height)
@@ -413,7 +429,7 @@ class MainWindow (WindowWithSavedPosition):
                 width, height = self.last_conf_size
                 sbwidth = lm.prefs[role].get("sbwidth", None)
                 if sbwidth is not None:
-                    handle_size = self.hpaned.style_get_property("handle-size")
+                    handle_size = style_get_property(self.hpaned, "handle-size")
                     pos = width - handle_size - sbwidth
                     self.hpaned.set_position(pos)
                     self.hpaned.queue_resize()
@@ -466,8 +482,16 @@ class SmallImageButton (gtk.Button):
 
     def on_style_set(self, widget, prev_style):
         settings = self.get_settings()
-        w, h = gtk.icon_size_lookup_for_settings(settings, gtk.ICON_SIZE_MENU)
-        self.set_size_request(w+4, h+4)
+        size_info = gtk.icon_size_lookup_for_settings(settings,
+                                                      gtk.ICON_SIZE_MENU)
+        if size_info:
+            if pygtkcompat.USE_GTK3:
+                valid, w, h = size_info
+            else:
+                w, h = size_info
+                valid = True
+            if valid:
+                self.set_size_request(w+4, h+4)
 
 
 class ToolResizeGrip (gtk.DrawingArea):
@@ -477,7 +501,7 @@ class ToolResizeGrip (gtk.DrawingArea):
     AREA_MIDDLE = 1
     AREA_RIGHT = 2
 
-    handle_size = gtk.HPaned().style_get_property("handle-size") + 2
+    handle_size = style_get_property(gtk.HPaned(), "handle-size") + 2
     corner_width = 4*handle_size
 
     window_edge_map = {
@@ -497,7 +521,12 @@ class ToolResizeGrip (gtk.DrawingArea):
         self.tool = tool
         self.set_size_request(4*self.corner_width, self.handle_size)
         self.connect("configure-event", self.on_configure_event)
-        self.connect("expose-event", self.on_expose_event)
+
+        if pygtkcompat.USE_GTK3:
+            self.connect("draw", self.on_draw)
+        else:
+            self.connect("expose-event", self.on_expose_event)
+
         self.connect("button-press-event", self.on_button_press_event)
         self.connect("button-release-event", self.on_button_release_event)
         self.connect("motion-notify-event", self.on_motion_notify_event)
@@ -512,17 +541,22 @@ class ToolResizeGrip (gtk.DrawingArea):
         self.width = event.width
         self.height = event.height
 
+    def on_draw(self, widget, cr):
+        # FIXME: implement
+        pass
+
     def on_expose_event(self, widget, event):
-        self.window.clear()
-        self.window.begin_paint_rect(event.area)
+        gdk_window = self.get_window()
+        gdk_window.clear()
+        gdk_window.begin_paint_rect(event.area)
         x = (self.width - self.handle_size) / 2
         y = (self.height - self.handle_size) / 2
         state = self.get_state()
-        self.style.paint_handle(self.window, state,
+        self.style.paint_handle(gdk_window, state,
             gtk.SHADOW_NONE, event.area, self, 'paned',
             0, 0, self.width, self.height,
             gtk.ORIENTATION_HORIZONTAL)
-        self.window.end_paint()
+        gdk_window.end_paint()
 
     def get_area(self, x, y):
         if self.tool.floating:
@@ -537,7 +571,7 @@ class ToolResizeGrip (gtk.DrawingArea):
             return
         if self.tool.floating:
             area = self.get_area(event.x, event.y)
-            win = self.tool.floating_window.window
+            win = self.tool.floating_window.get_window()
             edge = self.window_edge_map[area]
             win.begin_resize_drag(edge, event.button,
                                   int(event.x_root), int(event.y_root),
@@ -547,7 +581,7 @@ class ToolResizeGrip (gtk.DrawingArea):
             lm = self.tool.layout_manager
             max_w, max_h = lm.main_window.sidebar.max_tool_size()
             min_w = max_w
-            alloc = self.tool.allocation
+            alloc = self.tool.get_allocation()
             w = alloc.width
             h = alloc.height
             self.resize = event.x_root, event.y_root, w, h, \
@@ -557,7 +591,8 @@ class ToolResizeGrip (gtk.DrawingArea):
     def on_button_release_event(self, widget, event):
         self.resize = None
         self.grab_remove()
-        self.window.set_cursor(None)
+        gdk_window = self.get_window()
+        gdk_window.set_cursor(None)
 
     def get_cursor(self, area):
         if self.tool.floating:
@@ -567,15 +602,16 @@ class ToolResizeGrip (gtk.DrawingArea):
         return cursor
 
     def on_motion_notify_event(self, widget, event):
+        gdk_window = self.get_window()
         if not self.resize:
             area = self.get_area(event.x, event.y)
-            self.window.set_cursor(self.get_cursor(area))
+            gdk_window.set_cursor(self.get_cursor(area))
             return
         assert not self.tool.floating
         (ptr_x_root0, ptr_y_root0, w0, h0,
          min_w, min_h, max_w, max_h) = self.resize
         cursor = self.nonfloating_cursor
-        self.window.set_cursor(cursor)
+        gdk_window.set_cursor(cursor)
         dh = event.y_root - ptr_y_root0
         h = int(min(max(min_h, h0+dh), max_h))
         w = -1   # constrained horizontally anyway, better to not care
@@ -583,7 +619,8 @@ class ToolResizeGrip (gtk.DrawingArea):
         self.tool.queue_resize()
 
     def on_leave_notify_event(self, widget, event):
-        self.window.set_cursor(None)
+        gdk_window = self.get_window()
+        gdk_window.set_cursor(None)
 
 
 class FoldOutArrow (gtk.Button):
@@ -625,7 +662,8 @@ def draw_subtle_gradient(widget):
     light_rgba = _col2rgba(widget.style.light[state])
     mid_rgba = _col2rgba(widget.style.bg[state])
     dark_rgba = _col2rgba(widget.style.dark[state])
-    cr = widget.window.cairo_create()
+    gdk_window = widget.get_window()
+    cr = gdk_window.cairo_create()
     lg = cairo.LinearGradient(0, 0, 0, h)
     if state == gtk.STATE_ACTIVE:
         lg.add_color_stop_rgba(0.0, *dark_rgba)
@@ -646,10 +684,9 @@ class ToolDragHandle (gtk.EventBox):
     min_drag_distance = 10
     spacing = 2
 
-    def __init__(self, tool, stock_id):
+    def __init__(self, tool, stock_id, label_text):
         gtk.EventBox.__init__(self)
         self.stock_id = stock_id
-        stock_id, label_text, _mods, _key, _tr = gtk.stock_lookup(stock_id)
         self.tool = tool
         self.frame = frame = gtk.Frame()
         frame.set_shadow_type(gtk.SHADOW_OUT)
@@ -682,7 +719,11 @@ class ToolDragHandle (gtk.EventBox):
         self.in_reposition_drag = False
         # Floating status
         self.set_floating(False)
-        self.frame.connect("expose-event", self.on_frame_expose_event)
+
+        if pygtkcompat.USE_GTK3:
+            pass #FIXME: implement
+        else:
+            self.frame.connect("expose-event", self.on_frame_expose_event)
 
     def on_frame_expose_event(self, widget, event):
         draw_subtle_gradient(widget)
@@ -715,7 +756,8 @@ class ToolDragHandle (gtk.EventBox):
                 self._reset()
         elif event.type == gdk.BUTTON_PRESS:
             self.set_state(gtk.STATE_ACTIVE)
-            self.window.set_cursor(gdk.Cursor(gdk.FLEUR))
+            gdk_window = self.get_window()
+            gdk_window.set_cursor(gdk.Cursor(gdk.FLEUR))
             self.button_press_xy = event.x, event.y
 
     def on_button_release_event(self, widget, event):
@@ -724,8 +766,9 @@ class ToolDragHandle (gtk.EventBox):
     def _reset(self):
         self.button_press_xy = None
         self.set_state(gtk.STATE_NORMAL)
-        if self.window:
-            self.window.set_cursor(None)
+        gdk_window = self.get_window()
+        if gdk_window:
+            gdk_window.set_cursor(None)
 
     def on_motion_notify_event(self, widget, event):
         if not self.button_press_xy:
@@ -744,7 +787,8 @@ class ToolDragHandle (gtk.EventBox):
         lm = self.tool.layout_manager
         ix, iy = self.button_press_xy
         self.button_press_xy = None
-        self.window.set_cursor(gdk.Cursor(gdk.FLEUR))
+        gdk_window = self.get_window()
+        gdk_window.set_cursor(gdk.Cursor(gdk.FLEUR))
         self.set_state(gtk.STATE_ACTIVE)
         self.tool.layout_manager.drag_state.begin(self.tool, ix, iy)
 
@@ -753,11 +797,13 @@ class ToolDragHandle (gtk.EventBox):
         self._reset()
 
     def on_leave_notify_event(self, widget, event):
-        self.window.set_cursor(None)
+        gdk_window = self.get_window()
+        gdk_window.set_cursor(None)
         self.set_state(gtk.STATE_NORMAL)
 
     def on_enter_notify_event(self, widget, event):
-        self.window.set_cursor(gdk.Cursor(gdk.HAND2))
+        gdk_window = self.get_window()
+        gdk_window.set_cursor(gdk.Cursor(gdk.HAND2))
         #if not self.in_reposition_drag:
         #    self.set_state(gtk.STATE_PRELIGHT)
 
@@ -771,7 +817,12 @@ class ToolSnapBackBar (gtk.DrawingArea):
         gtk.DrawingArea.__init__(self)
         self.tool = tool
         self.set_size_request(-1, self.ARROW_SIZE)
-        self.connect("expose-event", self.on_expose_event)
+
+        if pygtkcompat.USE_GTK3:
+            pass #FIXME: implement
+        else:
+            self.connect("expose-event", self.on_expose_event)
+
         self.connect("button-press-event", self.on_button_down)
         self.connect("button-release-event", self.on_button_up)
         self.connect("enter-notify-event", self.on_enter)
@@ -823,7 +874,8 @@ class ToolSnapBackBar (gtk.DrawingArea):
         #    self, None, 0, 0, w, h)
         draw_subtle_gradient(widget)
 
-        sty.paint_arrow(self.window, state, gtk.SHADOW_IN, event.area,
+        gdk_window = self.get_window()
+        sty.paint_arrow(gdk_window, state, gtk.SHADOW_IN, event.area,
             self, "tearoffmenuitem", gtk.ARROW_RIGHT, True, w-h, 0, h, h)
 
 
@@ -832,7 +884,7 @@ class ToolSnapBackBar (gtk.DrawingArea):
         while x < right_max:
             x1 = x
             x2 = min(x+self.TEAR_LENGTH, right_max)
-            sty.paint_hline(self.window, gtk.STATE_NORMAL, event.area,
+            sty.paint_hline(gdk_window, gtk.STATE_NORMAL, event.area,
                        widget, "tearoffmenuitem",
                        x1, x2, (h - sty.ythickness)//2)
             x += 2 * self.TEAR_LENGTH
@@ -925,7 +977,7 @@ class ToolWindow (gtk.Window, ElasticContainer, WindowWithSavedPosition):
         x = max(x, 0)
         y = max(y, 0)
 
-        if self.window is not None:
+        if self.get_window() is not None:
             # It's mapped, so this will probably work even if the window
             # is not viewable right now.
             self.move(x, y)
@@ -954,7 +1006,7 @@ class Tool (gtk.VBox, ElasticContainer):
         self.role = role
         self.widget = widget
         self.layout_manager = layout_manager
-        self.handle = ToolDragHandle(self, stock_id)
+        self.handle = ToolDragHandle(self, stock_id, title)
         self.stock_id = stock_id
         self.snap_back_bar = ToolSnapBackBar(self)
         self.widget_frame = frame = gtk.Frame()
@@ -977,6 +1029,18 @@ class Tool (gtk.VBox, ElasticContainer):
         self.connect("style-set", self.on_style_set)
 
 
+    # Workaround for "widget" being an unsettable field in PyGI+GTK3.
+    # Hopefully this won't break anything and we won't need anything more.
+
+    @property
+    def widget(self):
+        return self._widget
+
+    @widget.setter
+    def widget(self, w):
+        self._widget = w
+
+
     def on_size_allocate(self, widget, allocation):
         if self.rolled_up:
             return
@@ -1000,7 +1064,10 @@ class Tool (gtk.VBox, ElasticContainer):
             if pixbuf is not None:
                 pixbufs.append(pixbuf)
         if pixbufs != []:
-            self.floating_window.set_icon_list(*pixbufs)
+            if pygtkcompat.USE_GTK3:
+                self.floating_window.set_icon_list(pixbufs)
+            else:
+                self.floating_window.set_icon_list(*pixbufs)
 
 
     def on_floating_window_delete_event(self, window, event):
@@ -1053,8 +1120,9 @@ class Tool (gtk.VBox, ElasticContainer):
                 wid.set_size_request(-1, -1)
                 wid.queue_resize()
 
-        if self.parent:
-            self.parent.remove(self)
+        parent = self.get_parent()
+        if parent is not None:
+            parent.remove(self)
         lm = self.layout_manager
         if lm.prefs.get(self.role, None) is None:
             lm.prefs[self.role] = {}
@@ -1113,8 +1181,9 @@ class Tool (gtk.VBox, ElasticContainer):
             lm.prefs[role] = {}
         if hidden:
             self.hide()
-            if self.parent:
-                self.parent.remove(self)
+            parent = self.get_parent()
+            if parent is not None:
+                parent.remove(self)
             if self.floating:
                 self.floating_window.hide()
         else:
@@ -1157,7 +1226,7 @@ class Tool (gtk.VBox, ElasticContainer):
         if resize_needed:
             self.queue_resize()
             while gtk.events_pending():
-                gtk.main_iteration(False)
+                gtk.main_iteration()
         # Notify the indicator arrow
         self.handle.set_rolled_up(rolled_up)
 
@@ -1167,14 +1236,15 @@ class Tool (gtk.VBox, ElasticContainer):
 
     def on_snap_button_pressed(self, window):
         # Mouse position
-        display = gdk.display_get_default()
+        display = pygtkcompat.gdk.display_get_default()
         screen, ptr_x, ptr_y, _modmask = display.get_pointer()
         if self.rolled_up:
             self.set_rolled_up(False)    # pending events will be processed
-        alloc = self.allocation
+        alloc = self.get_allocation()
         w = alloc.width
         h = alloc.height
-        titlebar_h = self.handle.allocation.height
+        handle_alloc = self.handle.get_allocation()
+        titlebar_h = handle_alloc.height
 
         def handle_snap():
             sidebar = self.layout_manager.main_window.sidebar
@@ -1215,22 +1285,32 @@ class ToolDragPreviewWindow (gtk.Window):
         self.set_decorated(False)
         self.set_position(gtk.WIN_POS_MOUSE)
         self.connect("map-event", self.on_map_event)
-        self.connect("expose-event", self.on_expose_event)
+
+        if pygtkcompat.USE_GTK3:
+            pass #FIXME: implement
+        else:
+            self.connect("expose-event", self.on_expose_event)
+
         self.connect("configure-event", self.on_configure_event)
         self.bg = None
         self.set_default_size(1,1)
 
     def show_all(self):
         gtk.Window.show_all(self)
-        self.window.move_resize(0, 0, 1, 1)
+        gdk_window = self.get_window()
+        gdk_window.move_resize(0, 0, 1, 1)
 
     def on_map_event(self, window, event):
         owner_win = self.owner.get_toplevel()
         self.set_transient_for(owner_win)
         # The first time we're mapped, set a background pixmap.
         # Background pixmap, a checkerboard
+        gdk_window = self.get_window()
         if self.bg is None:
-            self.bg = gdk.Pixmap(drawable=self.window, width=2, height=2)
+            if pygtkcompat.USE_GTK3:
+                # Until we have a non-Pixmap workaround.
+                return
+            self.bg = gdk.Pixmap(drawable=gdk_window, width=2, height=2)
             cmap = gdk.colormap_get_system()
             black = cmap.alloc_color(gdk.Color(0.0, 0.0, 0.0))
             white = cmap.alloc_color(gdk.Color(1.0, 1.0, 1.0))
@@ -1239,11 +1319,14 @@ class ToolDragPreviewWindow (gtk.Window):
             self.bg.draw_rectangle(gc, True, 0, 0, 2, 2)
             gc = self.bg.new_gc(white, black)
             self.bg.draw_points(gc, [(0,0), (1,1)])
-            self.window.set_back_pixmap(self.bg, False)
-            self.window.clear()
+            gdk_window.set_back_pixmap(self.bg, False)
+            gdk_window.clear()
 
     def on_configure_event(self, window, event):
         # Shape the window
+        if pygtkcompat.USE_GTK3:
+            # Until we have a non-Region workaround.
+            return
         w = event.width
         h = event.height
         r = gdk.Region()
@@ -1252,15 +1335,17 @@ class ToolDragPreviewWindow (gtk.Window):
         r.union_with_rect(gdk.Rectangle(0, 0, s, h))
         r.union_with_rect(gdk.Rectangle(0, h-s, w, s))
         r.union_with_rect(gdk.Rectangle(w-s, 0, s, h))
-        self.window.shape_combine_region(r, 0, 0)
+        gdk_window = self.get_window()
+        gdk_window.shape_combine_region(r, 0, 0)
 
     def on_expose_event(self, window, event):
         # Clear to the backing pixmap established earlier
         if self.bg is not None:
-            self.window.begin_paint_rect(event.area)
-            self.window.set_back_pixmap(self.bg, False)
-            self.window.clear()
-            self.window.end_paint()
+            gdk_window = self.get_window()
+            gdk_window.begin_paint_rect(event.area)
+            gdk_window.set_back_pixmap(self.bg, False)
+            gdk_window.clear()
+            gdk_window.end_paint()
 
 
 class ToolDragState:
@@ -1322,7 +1407,7 @@ class ToolDragState:
         # Establish a pointer grab on the main window (which stays mapped
         # throughout the drag).
         main_win = lm.main_window
-        main_win_gdk = main_win.window
+        main_win_gdk = main_win.get_window()
         events = gdk.BUTTON_PRESS_MASK | gdk.BUTTON_RELEASE_MASK \
             | gdk.BUTTON1_MOTION_MASK
         grab_status = gdk.pointer_grab(main_win_gdk, False, events,
@@ -1334,14 +1419,14 @@ class ToolDragState:
             self.tool = tool
             self.connect_reposition_handlers()
         except:
-            gdk.pointer_ungrab()
+            gdk.pointer_ungrab(gdk.CURRENT_TIME)
             self.disconnect_reposition_handlers()
             raise
 
     def on_reposition_button_release(self, widget, event):
         """Ends the current tool reposition drag.
         """
-        gdk.pointer_ungrab()
+        gdk.pointer_ungrab(gdk.CURRENT_TIME)
         self.disconnect_reposition_handlers()
         self.end()
 
@@ -1354,8 +1439,8 @@ class ToolDragState:
         # Moving out of the box defined by the sidebar
         # changes the mode of the current drag.
         sidebar = self.layout_manager.main_window.sidebar
-        sb_left, sb_top = sidebar.window.get_origin()
-        sb_alloc = sidebar.allocation
+        sb_left, sb_top = sidebar.get_window().get_origin()
+        sb_alloc = sidebar.get_allocation()
         sb_right = sb_left + sb_alloc.width
         sb_bottom = sb_top + sb_alloc.height
         if (x_root < sb_left or x_root > sb_right
@@ -1372,13 +1457,13 @@ class ToolDragState:
                 assert tool in sidebar.tools_vbox
                 sidebar.reorder_item(tool, ins)
                 sidebar.reassign_indices()
-            x, y = tool.handle.window.get_origin()
+            x, y = tool.handle.get_window().get_origin()
         else:
             ix, iy = self.handle_pos
             x = int(x_root - ix)
             y = int(y_root - iy)
         self.handle_pos_root = (x, y)
-        self.preview.window.move_resize(x, y, w, h)
+        self.preview.get_window().move_resize(x, y, w, h)
 
     def end(self):
         """Invoked at the end of repositioning. Snapping out happens here.
@@ -1407,7 +1492,7 @@ class Sidebar (gtk.EventBox):
     """Vertical sidebar containing reorderable tools which can be snapped out.
     """
 
-    MIN_SIZE = 150
+    MIN_SIZE = 230
 
     def __init__(self, layout_manager):
         gtk.EventBox.__init__(self)
@@ -1468,7 +1553,8 @@ class Sidebar (gtk.EventBox):
         Returns an integer position for passing to reorder_item(), or None.
         Currently only tool drag handles are valid insertion points.
         """
-        window_info = gdk.window_at_pointer()
+        display = self.get_display()
+        window_info = display.get_window_at_pointer()
         if window_info is None:
             return None
         pointer_window = window_info[0]
@@ -1476,12 +1562,12 @@ class Sidebar (gtk.EventBox):
         i = 0
         for widget in self.tools_vbox:
             if widget is self.slack:
-                if widget.window is pointer_window:
+                if widget.get_window() is pointer_window:
                     return i
             if isinstance(widget, Tool):
                 if widget is not current_tool:
                     dragger = widget.handle
-                    if dragger.window is pointer_window:
+                    if dragger.get_window() is pointer_window:
                         return i
             i += 1
         return None
@@ -1491,8 +1577,8 @@ class Sidebar (gtk.EventBox):
         """
         scrwin = self.scrolledwin
         viewpt = scrwin.get_child()
-        sb_pad = 2 * scrwin.style_get_property("scrollbar-spacing")
-        vp_alloc = viewpt.allocation
+        sb_pad = 2 * style_get_property(scrwin, "scrollbar-spacing")
+        vp_alloc = viewpt.get_allocation()
         max_size = (vp_alloc.width-sb_pad, vp_alloc.height-sb_pad)
         return max_size
 
@@ -1575,7 +1661,7 @@ def set_initial_window_position(win, pos):
     h = pos.get("h", None)
 
     # Where the mouse is right now
-    display = gdk.display_get_default()
+    display = pygtkcompat.gdk.display_get_default()
     screen, ptr_x, ptr_y, _modmask = display.get_pointer()
     if screen is None:
         raise RuntimeError, "No cursor on the default screen. Eek."
