@@ -10,22 +10,21 @@
 """
 
 import os
-from math import pi
 
+import pygtkcompat
 import gtk
 from gtk import gdk
 import gobject
 from gettext import gettext as _
 import pango
 
-from lib.helpers import hsv_to_rgb, clamp
 import dialogs
 from brushlib import brushsettings
-import stock
 import dropdownpanel
 import widgets
-from hsvcompat import ColorChangerHSV
-
+from colors import RGBColor, ColorAdjuster, HSVTriangle
+from colors import PreviousCurrentColorAdjuster, ColorPickerButton
+from lib.helpers import escape
 
 FRAMEWORK_XML = 'gui/toolbar.xml'
 MERGEABLE_XML = [
@@ -33,9 +32,13 @@ MERGEABLE_XML = [
     ("toolbar1_scrap", 'gui/toolbar-scrap.xml', _("Scraps switcher")),
     ("toolbar1_edit", 'gui/toolbar-edit.xml', _("Undo and Redo")),
     ("toolbar1_blendmodes", 'gui/toolbar-blendmodes.xml', _("Blend Modes")),
-    ("toolbar1_view", 'gui/toolbar-view.xml', _("View")),
+    ("toolbar1_linemodes", 'gui/toolbar-linemodes.xml', _("Line Mode")), #FIXME: make plural (string freeze hack)
+    ("toolbar1_view_modes", 'gui/toolbar-view-modes.xml', _("View (Main)")),
+    ("toolbar1_view_manips", 'gui/toolbar-view-manips.xml', _("View (Alternative/Secondary)")),
+    ("toolbar1_view_resets", 'gui/toolbar-view-resets.xml', _("View (Resetting)")),
     ("toolbar1_subwindows", 'gui/toolbar-subwindows.xml', _("Subwindows")),
     ]
+HISTORY_PREVIEW_SIZE = 48
 
 
 class ToolbarManager:
@@ -83,6 +86,8 @@ class ToolbarManager:
                     _("Current Brush"), None),
                 BrushSettingsDropdownToolAction("BrushSettingsDropdown", None,
                     _("Brush Settings"), None),
+                LineDropdownToolAction("LineDropdown", None,
+                    _("Line Mode"), None),
                 ]
         actions += self.item_actions
 
@@ -99,7 +104,8 @@ class ToolbarManager:
     def init_proxies(self):
         for action in self.item_actions:
             for p in action.get_proxies():
-                p.set_app(self.app)
+                if hasattr(p, "set_app"):
+                    p.set_app(self.app)
         # Merge in UI pieces based on the user's saved preferences
         for action in self.settings_actions:
             name = action.get_property("name")
@@ -109,9 +115,14 @@ class ToolbarManager:
 
     def on_toolbar1_popup_context_menu(self, toolbar, x, y, button):
         menu = self.toolbar1_popup
-        def posfunc(m):
+        def _posfunc(*a):
             return x, y, True
-        menu.popup(None, None, posfunc, button, 0)
+        time = gdk.CURRENT_TIME
+        data = None
+        if pygtkcompat.USE_GTK3:
+            menu.popup(None, None, _posfunc, data, button, time)
+        else:
+            menu.popup(None, None, _posfunc, button, time, data)
 
     def on_settings_toggle(self, toggleaction, ui_xml_file):
         name = toggleaction.get_property("name")
@@ -131,30 +142,152 @@ class ToolbarManager:
             self.toolbar1_ui_loaded.pop(name)
 
 
+
+from curve import CurveWidget
+
+class LineDropdownToolItem (gtk.ToolItem):
+    """Dropdown panel on the toolbar for changing line mode.
+    """
+
+    __gtype_name__ = "LineDropdownToolItem"
+    settings_coordinate = [('entry_pressure', (0,1)),
+                           ('midpoint_pressure', (1,1)),
+                           ('exit_pressure', (3,1)),
+                           ('line_head', (1,0)),
+                           ('line_tail', (2,0))]
+    action_names = ["SwitchableFreehandMode", "StraightMode",
+                    "SequenceMode", "EllipseMode"]
+
+
+    def __init__(self):
+        gtk.ToolItem.__init__(self)
+        self.set_homogeneous(False)
+        self.button_image = gtk.Image()
+        self.button_image.set_from_icon_name('mypaint-line-mode', ToolbarManager.icon_size)
+        self.line_mode_panel = dropdownpanel.DropdownPanelButton(self.button_image)
+        self.vbox = gtk.VBox()
+        self.vbox.set_border_width(widgets.SPACING_TIGHT)
+        self.vbox.set_spacing(widgets.SPACING_TIGHT)
+        frame = gtk.Frame()
+        frame.add(self.vbox)
+        frame.set_shadow_type(gtk.SHADOW_OUT)
+        self.line_mode_panel.set_property("panel-widget", frame)
+        self.add(self.line_mode_panel)
+        self.connect("create-menu-proxy", lambda *a: True)
+
+
+    def set_app(self, app):
+        """Deferred initialization: provides the main app instance.
+        """
+
+        self.app = app
+
+        # Action switcher buttons
+        bbox = gtk.HButtonBox()
+        frame = widgets.section_frame(_("Line Mode"))
+        frame.add(bbox)
+        bbox.set_border_width(widgets.SPACING)
+        for action_name in self.action_names:
+            action = app.find_action(action_name)
+            if action.get_active():
+                self.update_icon_from_action(action)
+            action.connect("changed", self.linemode_action_changed_cb)
+            button = gtk.ToggleButton()
+            action.connect_proxy(button)
+            button.connect("clicked", self.linemode_button_clicked_cb)
+            button.set_can_focus(False)
+            button.set_can_default(False)
+            button.set_image_position(gtk.POS_TOP)
+            button.set_relief(gtk.RELIEF_HALF)
+            image = action.create_icon(gtk.ICON_SIZE_LARGE_TOOLBAR)
+            image.set_padding(widgets.SPACING_TIGHT, widgets.SPACING_TIGHT)
+            button.set_image(image)
+            bbox.pack_start(button)
+        self.vbox.pack_start(frame, False, False)
+        bbox.show()
+
+        # Pressure settings.
+        frame = widgets.section_frame(_("Line Pressure"))
+        self.vbox.pack_start(frame, True, True)
+        curve = CurveWidget(npoints=4,
+                            ylockgroups=((1,2),),
+                            changed_cb=self.curve_changed_cb)
+        curve_align = gtk.Alignment(0, 0, 1, 1)
+        curve_align.add(curve)
+        curve_align.set_padding(widgets.SPACING, widgets.SPACING,
+                                widgets.SPACING, widgets.SPACING)
+        frame.add(curve_align)
+        curve_align.show()
+        curve.points = [(0.0,0.2), (0.33,.5),(0.66, .5), (1.0,.33)]
+        for setting, coord_pair in self.settings_coordinate:
+            adj = app.line_mode_settings.adjustments[setting]
+            value = adj.get_value()
+            index, subindex = coord_pair
+            if not setting.startswith ('line'):
+                value = 1.0 - value
+            coord = None
+            if subindex == 0:
+                coord = (value, curve.points[index][1])
+            else:
+                coord = (curve.points[index][0], value )
+            curve.set_point(index, coord)
+        self.curve_changed_cb (curve)
+
+
+    def update_icon_from_action(self, action):
+        """Updates the icon based on an action's icon.
+        """
+        icon_name = action.get_icon_name()
+        self.button_image.set_from_icon_name(icon_name, ToolbarManager.icon_size)
+
+
+    def linemode_action_changed_cb(self, action, current_action):
+        """Updates the dropdown button when the line mode changes."""
+        if action is current_action:
+            self.update_icon_from_action(action)
+
+
+    def linemode_button_clicked_cb(self, widget):
+        """Dismisses the dropdown panel when a linemode button is clicked.
+        """
+        gobject.idle_add(self.line_mode_panel.panel_hide)
+
+
+    def curve_changed_cb(self, curve):
+        """Updates the linemode pressure settings when the curve is altered.
+        """
+        for setting, coord_pair in self.settings_coordinate:
+            index, subindex = coord_pair
+            points = curve.points
+            value = curve.points[index][subindex]
+            if not setting.startswith('line'):
+                value = 1.0 - value
+            value = max(0.0001, value)
+            adj = self.app.line_mode_settings.adjustments[setting]
+            adj.set_value(value)
+
+
 class ColorDropdownToolItem (gtk.ToolItem):
     """Toolbar colour indicator, history access, and changer.
     """
 
     __gtype_name__ = "ColorDropdownToolItem"
 
-    HISTORY_PREVIEW_SIZE = 48
-
     def __init__(self, *a, **kw):
         gtk.ToolItem.__init__(self)
-        self.history_blobs = []
         self.main_blob = ColorBlob()
         self.dropdown_button = dropdownpanel.DropdownPanelButton(self.main_blob)
         self.app = None
         self.blob_size = ToolbarManager.icon_size
         self.connect("toolbar-reconfigured", self.on_toolbar_reconf)
-        self.connect("create-menu-proxy", self.on_create_menu_proxy)
+        self.connect("create-menu-proxy", lambda *a: True)
         self.set_tooltip_text(_("Color History and other tools"))
         self.add(self.dropdown_button)
 
     def set_app(self, app):
         self.app = app
         self.app.brush.observers.append(self.on_brush_settings_changed)
-        self.main_blob.hsv = self.app.brush.get_color_hsv()
+        self.main_blob.color = self.app.brush_color_manager.get_color()
 
         panel_frame = gtk.Frame()
         panel_frame.set_shadow_type(gtk.SHADOW_OUT)
@@ -179,19 +312,18 @@ class ColorDropdownToolItem (gtk.ToolItem):
         section_table.set_border_width(widgets.SPACING)
         section_frame.add(section_table)
 
-        hsv_widget = ColorChangerHSV(app, details=False)
+        hsv_widget = HSVTriangle()
         hsv_widget.set_size_request(175, 175)
+        hsv_widget.set_color_manager(app.brush_color_manager)
         section_table.attach(hsv_widget, 0, 1, 0, 1)
 
-        def is_preview_hbox(w):
-            return isinstance(w, gtk.HBox) and isinstance(w.parent, gtk.VBox)
-        preview_hbox, = widgets.find_widgets(hsv_widget, is_preview_hbox)
-        preview_hbox.parent.remove(preview_hbox)
-
-        def is_color_picker(w):
-            return isinstance(w, gtk.Button)
-        color_picker, = widgets.find_widgets(preview_hbox, is_color_picker)
-        color_picker.connect("clicked", hide_panel_idle_cb)
+        preview_hbox = gtk.HBox()
+        color_picker = ColorPickerButton()
+        preview_adj = PreviousCurrentColorAdjuster()
+        preview_adj.set_color_manager(app.brush_color_manager)
+        color_picker.set_color_manager(app.brush_color_manager)
+        preview_hbox.pack_start(color_picker, False, False)
+        preview_hbox.pack_start(preview_adj, True, True)
 
         side_vbox = gtk.VBox()
         side_vbox.set_spacing(widgets.SPACING_TIGHT)
@@ -199,7 +331,8 @@ class ColorDropdownToolItem (gtk.ToolItem):
 
         def init_proxy(widget, action_name):
             action = self.app.find_action(action_name)
-            action.connect_proxy(widget)
+            #action.connect_proxy(widget)  # deprecated since 2.16
+            widget.set_related_action(action)
             widget.connect("clicked", hide_panel_cb)
             return widget
 
@@ -208,55 +341,36 @@ class ColorDropdownToolItem (gtk.ToolItem):
         side_vbox.pack_end(preview_hbox, False, False)
 
         side_vbox.pack_end(gtk.Alignment(), True, True)
-        button = init_proxy(gtk.ToggleButton(), "ColorSamplerWindow")
-        side_vbox.pack_end(button, False, False)
-        button = init_proxy(gtk.ToggleButton(), "ColorSelectionWindow")
+        button = init_proxy(gtk.ToggleButton(), "ColorWindow")
+        button.set_label(_("Color Window"))
         side_vbox.pack_end(button, False, False)
 
         # History
         section_frame = widgets.section_frame(_("Recently Used"))
         panel_vbox.pack_start(section_frame, True, True)
 
-        self.history_blobs = []
-        history_hbox = gtk.HBox()
-        history_hbox.set_border_width(widgets.SPACING)
-        s = self.HISTORY_PREVIEW_SIZE
-        for hsv in app.ch.colors:
-            button = widgets.borderless_button()
-            blob = ColorBlob(hsv)
-            blob.set_size_request(s, s)
-            button.add(blob)
-            button.connect("clicked", self.on_history_button_clicked, blob)
-            history_hbox.pack_end(button, True, True)
-            self.history_blobs.append(blob)
-        app.ch.color_pushed_observers.append(self.color_pushed_cb)
-        section_frame.add(history_hbox)
+        history_view = ColorHistoryView(self)
+        section_frame.add(history_view)
 
-    def color_pushed_cb(self, color):
-        for blob, hsv in zip(self.history_blobs, self.app.ch.colors):
-            blob.hsv = hsv
-
-    def on_history_button_clicked(self, widget, blob):
-        self.app.brush.set_color_hsv(blob.hsv)
+    def on_history_button_clicked(self):
         self.dropdown_button.panel_hide()
 
-    def on_create_menu_proxy(self, toolitem):
-        # Do not appear on the overflow menu.
-        # Though possibly just duplicating the custom items into a submenu
-        # would work here.
-        self.set_proxy_menu_item("", None)
-        return True
-
     def on_toolbar_reconf(self, toolitem):
-        toolbar = self.parent
-        iw, ih = gtk.icon_size_lookup(self.get_icon_size())
+        toolbar = self.get_parent()
+        lookup_ret = gtk.icon_size_lookup(self.get_icon_size())
+        if pygtkcompat.USE_GTK3:
+            lookup_succeeded, iw, ih = lookup_ret
+            assert lookup_succeeded
+        else:
+            iw, ih = lookup_ret
         self.blob_size = max(iw, ih)
         self.main_blob.set_size_request(iw, ih)
 
     def on_brush_settings_changed(self, changes):
         if not changes.intersection(set(['color_h', 'color_s', 'color_v'])):
             return
-        self.main_blob.hsv = self.app.brush.get_color_hsv()
+        mgr = self.app.brush_color_manager
+        self.main_blob.color = mgr.get_color()
 
 
 
@@ -266,8 +380,6 @@ class BrushDropdownToolItem (gtk.ToolItem):
 
     __gtype_name__ = "BrushDropdownToolItem"
 
-    HISTORY_PREVIEW_SIZE = 48
-
     def __init__(self):
         gtk.ToolItem.__init__(self)
         self.history_images = []
@@ -276,7 +388,7 @@ class BrushDropdownToolItem (gtk.ToolItem):
         self.app = None
         self.image_size = ToolbarManager.icon_size
         self.connect("toolbar-reconfigured", self.on_toolbar_reconf)
-        self.connect("create-menu-proxy", self.on_create_menu_proxy)
+        self.connect("create-menu-proxy", lambda *a: True)
         self.set_tooltip_text(_("Brush history etc."))
         self.add(self.dropdown_button)
 
@@ -313,7 +425,7 @@ class BrushDropdownToolItem (gtk.ToolItem):
         # List editor button
         list_editor_button = gtk.ToggleButton()
         list_editor_action = self.app.find_action("BrushSelectionWindow")
-        list_editor_action.connect_proxy(list_editor_button)
+        list_editor_button.set_related_action(list_editor_action)
         close_panel_cb = lambda *a: self.dropdown_button.panel_hide()
         list_editor_button.connect("clicked", close_panel_cb)
         section_vbox.pack_start(list_editor_button, False, False)
@@ -332,11 +444,6 @@ class BrushDropdownToolItem (gtk.ToolItem):
             history_hbox.pack_end(button, True, True)
 
 
-    def on_create_menu_proxy(self, toolitem):
-        self.set_proxy_menu_item("", None)
-        return True
-
-
     def doc_input_stroke_ended_cb(self, event):
         gobject.idle_add(self.update_history_images)
 
@@ -344,7 +451,7 @@ class BrushDropdownToolItem (gtk.ToolItem):
     def update_history_images(self):
         bm = self.app.brushmanager
         if not self.history_images:
-            s = self.HISTORY_PREVIEW_SIZE
+            s = HISTORY_PREVIEW_SIZE
             for brush in bm.history:
                 image = ManagedBrushPreview()
                 image.set_size_request(s, s)
@@ -355,8 +462,13 @@ class BrushDropdownToolItem (gtk.ToolItem):
 
 
     def on_toolbar_reconf(self, toolitem):
-        toolbar = self.parent
-        iw, ih = gtk.icon_size_lookup(self.get_icon_size())
+        toolbar = self.get_parent()
+        lookup_ret = gtk.icon_size_lookup(self.get_icon_size())
+        if pygtkcompat.USE_GTK3:
+            lookup_succeeded, iw, ih = lookup_ret
+            assert lookup_succeeded
+        else:
+            iw, ih = lookup_ret
         self.image_size = max(iw, ih)
         self.main_image.set_size_request(iw, ih)
 
@@ -381,13 +493,22 @@ class BrushDropdownToolItem (gtk.ToolItem):
 class BrushSettingsDropdownToolItem (gtk.ToolItem):
     __gtype_name__ = "BrushSettingsDropdownToolItem"
 
+    active_stock_id = 'mypaint-brush-mods-active'
+    inactive_stock_id = 'mypaint-brush-mods-inactive'
+
     setting_cnames = ["radius_logarithmic", "slow_tracking", "opaque", "hardness"]
 
+    blend_modes_table = [
+        (0, 1, 0, 1, "BlendModeNormal"),
+        (0, 1, 1, 2, "BlendModeEraser"),
+        (1, 2, 0, 1, "BlendModeLockAlpha"),
+        (1, 2, 1, 2, "BlendModeColorize"),  ]
+
     def __init__(self):
         gtk.ToolItem.__init__(self)
         self.set_homogeneous(False)
         self.button_image = gtk.Image()
-        self.button_image.set_from_stock(stock.BRUSH_MODIFIERS_INACTIVE,
+        self.button_image.set_from_stock(self.inactive_stock_id,
                                          ToolbarManager.icon_size)
         self.button_shows_modified = False
         self.button = dropdownpanel.DropdownPanelButton(self.button_image)
@@ -397,11 +518,12 @@ class BrushSettingsDropdownToolItem (gtk.ToolItem):
         frame.set_shadow_type(gtk.SHADOW_OUT)
         self.button.set_property("panel-widget", frame)
         self.add(self.button)
+        self.connect("create-menu-proxy", lambda *a: True)
 
     def set_app(self, app):
         self.app = app
 
-        # A limited subset of the abailable brush settings.
+        # A limited subset of the available brush settings.
 
         frame = widgets.section_frame(_("Quick Brush Settings"))
         table = gtk.Table()
@@ -463,7 +585,7 @@ class BrushSettingsDropdownToolItem (gtk.ToolItem):
 
         widget = gtk.ToggleButton()
         action = self.app.find_action("BrushSettingsWindow")
-        action.connect_proxy(widget)
+        widget.set_related_action(action)
         #widget.set_label(_("Edit All Settings"))
         hbox.pack_start(widget, True, True)
         widget.connect("toggled", lambda a: self.button.panel_hide())
@@ -485,12 +607,26 @@ class BrushSettingsDropdownToolItem (gtk.ToolItem):
         frame.add(vbox)
         self.vbox.pack_start(frame, True, True)
 
-        for a in ["BlendModeNormal", "BlendModeEraser", "BlendModeLockAlpha"]:
-            action = self.app.find_action(a)
-            cb = gtk.CheckButton()
-            action.connect_proxy(cb)
-            vbox.pack_start(cb, False, False)
-            cb.connect("clicked", lambda a: self.button.panel_hide())
+        table = gtk.Table(2, 2, homogeneous=True)
+        topts = gtk.EXPAND|gtk.FILL
+        tpad = 0
+        table.set_row_spacings(widgets.SPACING_TIGHT)
+        table.set_col_spacings(widgets.SPACING)
+        for la, ra, ta, ba, action_name in self.blend_modes_table:
+            action = self.app.find_action(action_name)
+            button = gtk.ToggleButton()
+            action.connect_proxy(button)
+            button.set_can_focus(False)
+            button.set_can_default(False)
+            button.set_image_position(gtk.POS_LEFT)
+            button.set_alignment(0.0, 0.5)
+            button.set_relief(gtk.RELIEF_HALF)
+            image = action.create_icon(gtk.ICON_SIZE_BUTTON)
+            image.set_padding(widgets.SPACING_TIGHT, widgets.SPACING_TIGHT)
+            button.set_image(image)
+            table.attach(button, la, ra, ta, ba, topts, topts, tpad, tpad)
+            button.connect("clicked", self.blendmode_button_clicked_cb)
+        vbox.pack_start(table, False, False)
 
         self.vbox.set_border_width(widgets.SPACING)
         self.vbox.set_spacing(widgets.SPACING)
@@ -498,6 +634,10 @@ class BrushSettingsDropdownToolItem (gtk.ToolItem):
         self.app.brush.observers.append(self.brush_settings_changed_cb)
 
 
+    def blendmode_button_clicked_cb(self, widget):
+        self.button.panel_hide()
+
+
     def reset_button_clicked_cb(self, widget, adj, setting_cname):
         default = self._get_current_brush_default(setting_cname)
         adj.set_value(default)
@@ -529,13 +669,13 @@ class BrushSettingsDropdownToolItem (gtk.ToolItem):
         stock_id = None
         if self._current_brush_is_modified():
             if not self.button_shows_modified:
-                stock_id = stock.BRUSH_MODIFIERS_ACTIVE
+                stock_id = self.active_stock_id
                 self.button_shows_modified = True
             if not self.reset_all_button.get_sensitive():
                 self.reset_all_button.set_sensitive(True)
         else:
             if self.button_shows_modified:
-                stock_id = stock.BRUSH_MODIFIERS_INACTIVE
+                stock_id = self.inactive_stock_id
                 self.button_shows_modified = False
             if self.reset_all_button.get_sensitive():
                 self.reset_all_button.set_sensitive(False)
@@ -553,24 +693,33 @@ class BrushSettingsDropdownToolItem (gtk.ToolItem):
 
 class ColorDropdownToolAction (gtk.Action):
     __gtype_name__ = "ColorDropdownToolAction"
-ColorDropdownToolAction.set_tool_item_type(ColorDropdownToolItem)
+    def do_create_tool_item(self):
+        return ColorDropdownToolItem()
+
+
+class LineDropdownToolAction (gtk.Action):
+    __gtype_name__ = "LineDropdownToolAction"
+    def do_create_tool_item(self):
+        return LineDropdownToolItem()
 
 
 class BrushDropdownToolAction (gtk.Action):
     __gtype_name__ = "BrushDropdownToolAction"
-BrushDropdownToolAction.set_tool_item_type(BrushDropdownToolItem)
+    def do_create_tool_item(self):
+        return BrushDropdownToolItem()
 
 
 class BrushSettingsDropdownToolAction (gtk.Action):
     __gtype_name__ = "BrushSettingsDropdownToolAction"
-BrushSettingsDropdownToolAction\
-        .set_tool_item_type(BrushSettingsDropdownToolItem)
+    def do_create_tool_item(self):
+        return BrushSettingsDropdownToolItem()
 
 
 class ManagedBrushPreview (gtk.Image):
     """Updateable widget displaying a brushmanager.ManagedBrush`'s preview.
     """
 
+    ICON_SIZE = 32
     TOOLTIP_ICON_SIZE = 48
 
     def __init__(self, brush=None):
@@ -579,7 +728,8 @@ class ManagedBrushPreview (gtk.Image):
         self.image_size = None
         self.brush_name = None
         self.set_from_managed_brush(brush)
-        self.set_size_request(32, 32)
+        s = self.ICON_SIZE
+        self.set_size_request(s, s)
         self.connect("size-allocate", self.on_size_allocate)
         self.connect("query-tooltip", self.on_query_tooltip)
         self.set_property("has-tooltip", True)
@@ -595,7 +745,12 @@ class ManagedBrushPreview (gtk.Image):
 
     def on_size_allocate(self, widget, alloc):
         new_size = alloc.width, alloc.height
-        if new_size != self.image_size:
+        # if new_size != self.image_size:
+        if self.image_size is None:
+            # XXX dubious fix: what if the preview receives a new size in the
+            # middle of its lifetime? Under GTK3 however, permitting this makes
+            # the preview keep growing by about 4px each penstroke or brush
+            # selection. Not sure why.
             self.image_size = alloc.width, alloc.height
             self._update()
 
@@ -610,7 +765,7 @@ class ManagedBrushPreview (gtk.Image):
         s = self.TOOLTIP_ICON_SIZE
         scaled_pixbuf = self._get_scaled_pixbuf(s)
         tooltip.set_icon(scaled_pixbuf)
-        tooltip.set_text(self.brush_name)  # XXX markup and summary of changes
+        tooltip.set_text(self.brush_name)  # TODO: use markup, and summarize changes (i18n fun)
         return True
 
     def _update(self):
@@ -622,35 +777,74 @@ class ManagedBrushPreview (gtk.Image):
         self.set_from_pixbuf(scaled_pixbuf)
 
 
+class ColorHistoryView (gtk.HBox, ColorAdjuster):
+    """A set of ColorBlobs showing the usage history.
+    """
+
+    def __init__(self, toolitem):
+        gtk.HBox.__init__(self)
+        self.__history_blobs = []
+        self.__app = toolitem.app
+        self.__toolitem = toolitem
+        self.set_border_width(widgets.SPACING)
+        s = HISTORY_PREVIEW_SIZE
+        mgr = self.__app.brush_color_manager
+        for color in mgr.get_history():
+            button = widgets.borderless_button()
+            blob = ColorBlob(color)
+            blob.set_size_request(s, s)
+            button.add(blob)
+            button.connect("clicked", self.__button_clicked_cb)
+            self.pack_end(button, True, True)
+            self.__history_blobs.append(blob)
+        self.set_color_manager(mgr)
+
+    def update_cb(self):
+        mgr = self.get_color_manager()
+        for blob, color in zip(self.__history_blobs, mgr.get_history()):
+            blob.color = color
+
+    def __button_clicked_cb(self, button):
+        blob = button.get_child()
+        mgr = self.get_color_manager()
+        mgr.set_color(blob.color)
+        self.__toolitem.on_history_button_clicked()
+
 
 class ColorBlob (gtk.AspectFrame):
     """Updatable widget displaying a single colour.
     """
 
-    def __init__(self, hsv=None):
+    def __init__(self, color=None):
         gtk.AspectFrame.__init__(self, xalign=0.5, yalign=0.5, ratio=1.0, obey_child=False)
         self.set_name("thinborder-color-blob-%d" % id(self))
         self.set_shadow_type(gtk.SHADOW_IN)
         self.drawingarea = gtk.DrawingArea()
         self.add(self.drawingarea)
-        if hsv is None:
-            hsv = 0.0, 0.0, 0.0
-        self._hsv = hsv
+        if color is None:
+            color = RGBColor(0, 0, 0)
+        self._color = color
         self.drawingarea.set_size_request(1, 1)
-        self.drawingarea.connect("expose-event", self.on_expose)
+        if pygtkcompat.USE_GTK3:
+            self.drawingarea.connect("draw", self.on_draw)
+        else:
+            self.drawingarea.connect("expose-event", self.on_expose)
 
-    def set_hsv(self, hsv):
-        self._hsv = hsv
+    def set_color(self, color):
+        self._color = color
         self.drawingarea.queue_draw()
 
-    def get_hsv(self):
-        return self._hsv
+    def get_color(self):
+        return self._color
 
-    hsv = property(get_hsv, set_hsv)
+    color = property(get_color, set_color)
 
     def on_expose(self, widget, event):
-        cr = widget.window.cairo_create()
-        cr.set_source_rgb(*hsv_to_rgb(*self._hsv))
+        cr = widget.get_window().cairo_create()
+        self.on_draw(widget, cr)
+
+    def on_draw(self, widget, cr):
+        cr.set_source_rgb(*self._color.get_rgb())
         cr.paint()
 
 
@@ -673,20 +867,20 @@ class MainMenuButton (gtk.ToggleButton):
         self.menu = menu
         hbox1 = gtk.HBox()
         hbox2 = gtk.HBox()
-        label = gtk.Label(text)
+        label = gtk.Label()
         hbox1.pack_start(label, True, True)
         arrow = gtk.Arrow(gtk.ARROW_DOWN, gtk.SHADOW_IN)
         hbox1.pack_start(arrow, False, False)
         hbox2.pack_start(hbox1, True, True, widgets.SPACING_TIGHT)
 
         # Text settings
-        attrs = pango.AttrList()
-        attrs.change(pango.AttrWeight(pango.WEIGHT_SEMIBOLD, 0, -1))
-        label.set_attributes(attrs)
+        text = unicode(text)
+        markup = "<b>%s</b>" % escape(text)
+        label.set_markup(markup)
 
         self.add(hbox2)
         self.set_relief(gtk.RELIEF_NONE)
-        self.set_can_focus(True)
+        self.set_can_focus(False)
         self.set_can_default(False)
         self.connect("toggled", self.on_toggled)
 
@@ -719,7 +913,7 @@ class MainMenuButton (gtk.ToggleButton):
         if togglebutton.get_active():
             if not self.menu.get_property("visible"):
                 pos_func = self._get_popup_menu_position
-                self.menu.popup(None, None, pos_func, 1, 0)
+                pygtkcompat.gtk.menu_popup(self.menu, None, None, pos_func, 1, 0)
 
 
     def on_menu_dismiss(self, *a, **kw):
@@ -732,6 +926,6 @@ class MainMenuButton (gtk.ToggleButton):
 
     def _get_popup_menu_position(self, menu, *junk):
         # Underneath the button, at the same x position.
-        x, y = self.window.get_origin()
-        y += self.allocation.height
+        x, y = self.get_window().get_origin()
+        y += self.get_allocation().height
         return x, y, True
