@@ -15,7 +15,13 @@ class CommandStack:
         self.call_before_action = []
         self.stack_observers = []
         self.clear()
-    
+
+    def __repr__(self):
+        return "<CommandStack\n  <Undo len=%d last3=%r>\n" \
+                "  <Redo len=%d last3=%r> >" % (
+                    len(self.undo_stack), self.undo_stack[-3:],
+                    len(self.redo_stack), self.redo_stack[:3],  )
+
     def clear(self):
         self.undo_stack = []
         self.redo_stack = []
@@ -28,7 +34,7 @@ class CommandStack:
         self.undo_stack.append(command)
         self.reduce_undo_history()
         self.notify_stack_observers()
-    
+
     def undo(self):
         if not self.undo_stack: return
         for f in self.call_before_action: f()
@@ -37,7 +43,7 @@ class CommandStack:
         self.redo_stack.append(command)
         self.notify_stack_observers()
         return command
-        
+
     def redo(self):
         if not self.redo_stack: return
         for f in self.call_before_action: f()
@@ -62,22 +68,75 @@ class CommandStack:
         if not self.undo_stack: return None
         return self.undo_stack[-1]
 
+    def update_last_command(self, **kwargs):
+        cmd = self.get_last_command()
+        if cmd is None:
+            return None
+        cmd.update(**kwargs)
+        self.notify_stack_observers() # the display_name may have changed
+        return cmd
+
     def notify_stack_observers(self):
         for func in self.stack_observers:
             func(self)
 
 class Action:
-    '''Base class for all undo/redoable actions. Subclasses must implement the
+    """An undoable, redoable action.
+
+    Base class for all undo/redoable actions. Subclasses must implement the
     undo and redo methods. They should have a reference to the document in 
-    self.doc'''
+    self.doc.
+
+    """
     automatic_undo = False
     display_name = _("Unknown Action")
 
+    def __repr__(self):
+        return "<%s>" % (self.display_name,)
+
+
     def redo(self):
+        """Callback used to perform, or re-perform the Action.
+        """
         raise NotImplementedError
+
+
     def undo(self):
+        """Callback used to un-perform an already performed Action.
+        """
         raise NotImplementedError
 
+
+    def update(self, **kwargs):
+        """In-place update on the tip of the undo stack.
+
+        This method should update the model in the way specified in `**kwargs`.
+        The interpretation of arguments is left to the concrete implementation.
+
+        Updating the top Action on the command stack is used to prevent
+        situations where an undo() followed by a redo() would result in
+        multiple sendings of GdkEvents by code designed to keep interface state
+        in sync with the model.
+
+        """
+
+        # Updating is used in situations where only the user's final choice of
+        # a state such as layer visibility matters in the command-stream.
+        # Creating a nice workflow for the user by using `undo()` then `do()`
+        # with a replacement Action can sometimes cause GtkAction and
+        # command.Action flurries or loops across multiple GdkEvent callbacks.
+        #
+        # This can make coding difficult elsewhere. For example,
+        # GtkToggleActions must be kept in in sync with undoable boolean model
+        # state, but even when an interlock or check is coded, the fact that
+        # processing happens in multiple GtkEvent handlers can result in,
+        # essentially, a toggle action which turns itself off immediately after
+        # being toggled on. See https://gna.org/bugs/?20096 for a concrete
+        # example.
+
+        raise NotImplementedError
+
+
     # Utility functions
     def _notify_canvas_observers(self, affected_layers):
         bbox = helpers.Rect()
@@ -138,8 +197,12 @@ class MergeLayer(Action):
     def __init__(self, doc, dst_idx):
         self.doc = doc
         self.dst_layer = self.doc.layers[dst_idx]
+        self.normalize_src = ConvertLayerToNormalMode(doc, doc.layer)
+        self.normalize_dst = ConvertLayerToNormalMode(doc, self.dst_layer)
         self.remove_src = RemoveLayer(doc)
     def redo(self):
+        self.normalize_src.redo()
+        self.normalize_dst.redo()
         self.dst_before = self.dst_layer.save_snapshot()
         assert self.doc.layer is not self.dst_layer
         self.doc.layer.merge_into(self.dst_layer)
@@ -153,8 +216,32 @@ class MergeLayer(Action):
         self.remove_src.undo()
         self.dst_layer.load_snapshot(self.dst_before)
         del self.dst_before
+        self.normalize_dst.undo()
+        self.normalize_src.undo()
         self._notify_document_observers()
 
+class ConvertLayerToNormalMode(Action):
+    display_name = _("Convert Layer Mode")
+    def __init__(self, doc, layer):
+        self.doc = doc
+        self.layer = layer
+        self.set_normal_mode = SetLayerCompositeOp(doc, 'svg:src-over', layer)
+        self.set_opacity = SetLayerOpacity(doc, 1.0, layer)
+    def redo(self):
+        self.before = self.layer.save_snapshot()
+        prev_idx = self.doc.layer_idx
+        self.doc.layer_idx = self.doc.layers.index(self.layer)
+        get_bg = self.doc.get_rendered_image_behind_current_layer
+        self.layer.convert_to_normal_mode(get_bg)
+        self.doc.layer_idx = prev_idx
+        self.set_normal_mode.redo()
+        self.set_opacity.redo()
+    def undo(self):
+        self.set_opacity.undo()
+        self.set_normal_mode.undo()
+        self.layer.load_snapshot(self.before)
+        del self.before
+
 class AddLayer(Action):
     display_name = _("Add Layer")
     def __init__(self, doc, insert_idx=None, after=None, name=''):
@@ -165,6 +252,7 @@ class AddLayer(Action):
             self.insert_idx = l_idx + 1
         self.layer = layer.Layer(name)
         self.layer.content_observers.append(self.doc.layer_modified_cb)
+        self.layer.set_symmetry_axis(self.doc.get_symmetry_axis())
     def redo(self):
         self.doc.layers.insert(self.insert_idx, self.layer)
         self.prev_idx = self.doc.layer_idx
@@ -176,23 +264,37 @@ class AddLayer(Action):
         self._notify_document_observers()
 
 class RemoveLayer(Action):
+    """Removes a layer, replacing it with a new one if it was the last.
+    """
     display_name = _("Remove Layer")
     def __init__(self, doc,layer=None):
         self.doc = doc
         self.layer = layer
+        self.newlayer0 = None
     def redo(self):
-        assert len(self.doc.layers) > 1
         if self.layer:
             self.idx = self.doc.layers.index(self.layer)
             self.doc.layers.remove(self.layer)
         else:
             self.idx = self.doc.layer_idx
             self.layer = self.doc.layers.pop(self.doc.layer_idx)
-        if self.doc.layer_idx == len(self.doc.layers):
-            self.doc.layer_idx -= 1
+        if len(self.doc.layers) == 0:
+            if self.newlayer0 is None:
+                ly = layer.Layer("")
+                ly.content_observers.append(self.doc.layer_modified_cb)
+                ly.set_symmetry_axis(self.doc.get_symmetry_axis())
+                self.newlayer0 = ly
+            self.doc.layers.append(self.newlayer0)
+            self.doc.layer_idx = 0
+            assert self.idx == 0
+        else:
+            if self.doc.layer_idx == len(self.doc.layers):
+                self.doc.layer_idx -= 1
         self._notify_canvas_observers([self.layer])
         self._notify_document_observers()
     def undo(self):
+        if self.newlayer0 is not None:
+            self.doc.layers.remove(self.newlayer0)
         self.doc.layers.insert(self.idx, self.layer)
         self.doc.layer_idx = self.idx
         self._notify_canvas_observers([self.layer])
@@ -214,7 +316,32 @@ class SelectLayer(Action):
         self._notify_document_observers()
 
 class MoveLayer(Action):
-    display_name = _("Move Layer")
+    display_name = _("Move Layer on Canvas")
+    # NOT "Move Layer" for now - old translatable string with different sense
+    def __init__(self, doc, layer_idx, dx, dy, ignore_first_redo=True):
+        self.doc = doc
+        self.layer_idx = layer_idx
+        self.dx = dx
+        self.dy = dy
+        self.ignore_first_redo = ignore_first_redo
+    def redo(self):
+        layer = self.doc.layers[self.layer_idx]
+        if self.ignore_first_redo:
+            # these are typically created interactively, after
+            # the entire layer has been moved
+            self.ignore_first_redo = False
+        else:
+            layer.translate(self.dx, self.dy)
+        self._notify_canvas_observers([layer])
+        self._notify_document_observers()
+    def undo(self):
+        layer = self.doc.layers[self.layer_idx]
+        layer.translate(-self.dx, -self.dy)
+        self._notify_canvas_observers([layer])
+        self._notify_document_observers()
+
+class ReorderSingleLayer(Action):
+    display_name = _("Reorder Layer in Stack")
     def __init__(self, doc, was_idx, new_idx, select_new=False):
         self.doc = doc
         self.was_idx = was_idx
@@ -248,6 +375,7 @@ class DuplicateLayer(Action):
         self.new_layer = layer.Layer(name)
         self.new_layer.load_snapshot(snapshot)
         self.new_layer.content_observers.append(self.doc.layer_modified_cb)
+        self.new_layer.set_symmetry_axis(doc.get_symmetry_axis())
     def redo(self):
         self.doc.layers.insert(self.insert_idx+1, self.new_layer)
         self.duplicate_layer = self.doc.layers[self.insert_idx+1]
@@ -260,7 +388,7 @@ class DuplicateLayer(Action):
         self._notify_document_observers()
 
 class ReorderLayers(Action):
-    display_name = _("Reorder Layer")
+    display_name = _("Reorder Layer Stack")
     def __init__(self, doc, new_order):
         self.doc = doc
         self.old_order = doc.layers[:]
@@ -280,6 +408,20 @@ class ReorderLayers(Action):
         self._notify_canvas_observers(self.doc.layers)
         self._notify_document_observers()
 
+class RenameLayer(Action):
+    display_name = _("Rename Layer")
+    def __init__(self, doc, name, layer):
+        self.doc = doc
+        self.new_name = name
+        self.layer = layer
+    def redo(self):
+        self.old_name = self.layer.name
+        self.layer.name = self.new_name
+        self._notify_document_observers()
+    def undo(self):
+        self.layer.name = self.old_name
+        self._notify_document_observers()
+
 class SetLayerVisibility(Action):
     def __init__(self, doc, visible, layer):
         self.doc = doc
@@ -294,6 +436,11 @@ class SetLayerVisibility(Action):
         self.layer.visible = self.old_visibility
         self._notify_canvas_observers([self.layer])
         self._notify_document_observers()
+    def update(self, visible):
+        self.layer.visible = visible
+        self.new_visibility = visible
+        self._notify_canvas_observers([self.layer])
+        self._notify_document_observers()
     @property
     def display_name(self):
         if self.new_visibility:
@@ -315,6 +462,11 @@ class SetLayerLocked (Action):
         self.layer.locked = self.old_locked
         self._notify_canvas_observers([self.layer])
         self._notify_document_observers()
+    def update(self, locked):
+        self.layer.locked = locked
+        self.new_locked = locked
+        self._notify_canvas_observers([self.layer])
+        self._notify_document_observers()
     @property
     def display_name(self):
         if self.new_locked:
@@ -333,17 +485,21 @@ class SetLayerOpacity(Action):
             l = self.layer
         else:
             l = self.doc.layer
+        previous_effective_opacity = l.effective_opacity
         self.old_opacity = l.opacity
         l.opacity = self.new_opacity
-        self._notify_canvas_observers([l])
+        if l.effective_opacity != previous_effective_opacity:
+            self._notify_canvas_observers([l])
         self._notify_document_observers()
     def undo(self):
         if self.layer:
             l = self.layer
         else:
             l = self.doc.layer
+        previous_effective_opacity = l.effective_opacity
         l.opacity = self.old_opacity
-        self._notify_canvas_observers([l])
+        if l.effective_opacity != previous_effective_opacity:
+            self._notify_canvas_observers([l])
         self._notify_document_observers()
 
 class SetLayerCompositeOp(Action):
