@@ -6,7 +6,11 @@
 # the Free Software Foundation; either version 2 of the License, or
 # (at your option) any later version.
 
-import os, zipfile, tempfile, time, traceback
+import gui.pygtkcompat
+if gui.pygtkcompat.USE_GTK3:
+    from gi.repository import GdkPixbuf
+
+import os, sys, zipfile, tempfile, time, traceback
 join = os.path.join
 from cStringIO import StringIO
 import xml.etree.ElementTree as ET
@@ -23,10 +27,16 @@ LOAD_CHUNK_SIZE = 64*1024
 
 from layer import DEFAULT_COMPOSITE_OP, VALID_COMPOSITE_OPS
 
+
 class SaveLoadError(Exception):
-    """Expected errors on loading or saving, like missing permissions or non-existing files."""
+    """Expected errors on loading or saving
+
+    Covers stuff like missing permissions or non-existing files.
+
+    """
     pass
 
+
 class Document():
     """
     This is the "model" in the Model-View-Controller design.
@@ -48,57 +58,36 @@ class Document():
         if not brushinfo:
             brushinfo = brush.BrushInfo()
             brushinfo.load_defaults()
+        self.layers = []
         self.brush = brush.Brush(brushinfo)
+        self.brush.brushinfo.observers.append(self.brushsettings_changed_cb)
         self.stroke = None
-        self.canvas_observers = []
-        self.stroke_observers = [] # callback arguments: stroke, brush (brush is a temporary read-only convenience object)
-        self.doc_observers = []
+        self.canvas_observers = []  #: See `layer_modified_cb()`
+        self.stroke_observers = [] #: See `split_stroke()`
+        self.doc_observers = [] #: See `call_doc_observers()`
         self.frame_observers = []
         self.command_stack_observers = []
+        self.symmetry_observers = []  #: See `set_symmetry_axis()`
+        self.__symmetry_axis = None
+        self.default_background = (255, 255, 255)
         self.clear(True)
 
         self._frame = [0, 0, 0, 0]
         self._frame_enabled = False
-        # Used by move_frame() to accumulate values
-        self._frame_dx = 0.0
-        self._frame_dy = 0.0
+
+    def move_current_layer(self, dx, dy):
+        layer = self.layers[self.layer_idx]
+        layer.translate(dx, dy)
 
     def get_frame(self):
         return self._frame
 
-    def move_frame(self, dx=0.0, dy=0.0):
-        """Move the frame. Accumulates changes and moves the frame once
-        the accumulated change reaches the minimum move step."""
-        # FIXME: Should be 1 (pixel aligned), not tile aligned
-        # This is due to PNG saving having to be tile aligned
-        min_step = N
-
-        def round_to_n(value, n):
-            return int(round(value/n)*n)
-
-        x, y, w, h = self.get_frame()
-
-        self._frame_dx += dx
-        self._frame_dy += dy
-        step_x = round_to_n(self._frame_dx, min_step)
-        step_y = round_to_n(self._frame_dy, min_step)
-
-        if step_x:
-            self.set_frame(x=x+step_x)
-            self._frame_dx -= step_x
-
-        if step_y:
-            self.set_frame(y=y+step_y)
-            self._frame_dy -= step_y
-
     def set_frame(self, x=None, y=None, width=None, height=None):
         """Set the size of the frame. Pass None to indicate no-change."""
 
         for i, var in enumerate([x, y, width, height]):
             if not var is None:
-                # FIXME: must be aligned to tile size due to PNG saving
-                assert not var % N, "Frame size must be aligned to tile size"
-                self._frame[i] = var
+                self._frame[i] = int(var)
 
         for f in self.frame_observers: f()
 
@@ -110,20 +99,50 @@ class Document():
         for f in self.frame_observers: f()
     frame_enabled = property(get_frame_enabled)
 
+
     def call_doc_observers(self):
+        """Announce major structural changes via `doc_observers`.
+
+        This is invoked to announce major structural changes such as the layers
+        changing or a new document being loaded. The callbacks in the list are
+        invoked with a single argument, `self`.
+
+        """
         for f in self.doc_observers:
             f(self)
         return True
 
+
+    def get_symmetry_axis(self):
+        """Gets the active painting symmetry X axis value.
+        """
+        return self.__symmetry_axis
+
+
+    def set_symmetry_axis(self, x):
+        """Sets the active painting symmetry X axis value.
+
+        A value of `None` inactivates symmetrical painting. After setting, all
+        registered `symmetry_observers` are called without arguments.
+        """
+        # TODO: make this undoable?
+        for layer in self.layers:
+            layer.set_symmetry_axis(x)
+        self.__symmetry_axis = x
+        for func in self.symmetry_observers:
+            func()
+
+
     def clear(self, init=False):
         self.split_stroke()
+        self.set_symmetry_axis(None)
         if not init:
             bbox = self.get_bbox()
         # throw everything away, including undo stack
 
         self.command_stack = command.CommandStack()
         self.command_stack.stack_observers = self.command_stack_observers
-        self.set_background((255, 255, 255))
+        self.set_background(self.default_background)
         self.layers = []
         self.layer_idx = None
         self.add_layer(0)
@@ -141,22 +160,52 @@ class Document():
         return self.layers[self.layer_idx]
     layer = property(get_current_layer)
 
+
     def split_stroke(self):
+        """Splits the current stroke, announcing the newly stacked stroke
+
+        The stroke being drawn is pushed onto to the command stack and the
+        callbacks in the list `self.stroke_observers` are invoked with two
+        arguments: the newly completed stroke, and the brush used. The brush
+        argument is a temporary read-only convenience object.
+
+        This is called every so often when drawing a single long brushstroke on
+        input to allow parts of a long line to be undone.
+
+        """
         if not self.stroke: return
         self.stroke.stop_recording()
         if not self.stroke.empty:
-            self.command_stack.do(command.Stroke(self, self.stroke, self.snapshot_before_stroke))
+            cmd = command.Stroke(self, self.stroke,
+                                 self.snapshot_before_stroke)
+            self.command_stack.do(cmd)
             del self.snapshot_before_stroke
             self.unsaved_painting_time += self.stroke.total_painting_time
             for f in self.stroke_observers:
                 f(self.stroke, self.brush)
         self.stroke = None
 
+
+    def brushsettings_changed_cb(self, settings, lightweight_settings=set([
+            'radius_logarithmic', 'color_h', 'color_s', 'color_v',
+            'opaque', 'hardness', 'slow_tracking', 'slow_tracking_per_dab'
+            ])):
+        # The lightweight brush settings are expected to change often in
+        # mid-stroke e.g. by heavy keyboard usage. If only those change, we
+        # don't create a new undo step. (And thus also no separate pickable
+        # stroke in the strokemap.)
+        if settings - lightweight_settings:
+            self.split_stroke()
+
     def select_layer(self, idx):
         self.do(command.SelectLayer(self, idx))
 
+    def record_layer_move(self, layer, dx, dy):
+        layer_idx = self.layers.index(layer)
+        self.do(command.MoveLayer(self, layer_idx, dx, dy, True))
+
     def move_layer(self, was_idx, new_idx, select_new=False):
-        self.do(command.MoveLayer(self, was_idx, new_idx, select_new))
+        self.do(command.ReorderSingleLayer(self, was_idx, new_idx, select_new))
 
     def duplicate_layer(self, insert_idx=None, name=''):
         self.do(command.DuplicateLayer(self, insert_idx, name))
@@ -168,7 +217,31 @@ class Document():
         if not self.layer.is_empty():
             self.do(command.ClearLayer(self))
 
+
     def stroke_to(self, dtime, x, y, pressure, xtilt, ytilt):
+        """Draws a stroke to the current layer with the current brush.
+
+        This is called by GUI code in response to motion events on the canvas -
+        both with and without pressure. If enough time has elapsed,
+        `split_stroke()` is called.
+
+        :param self:
+            This is an object method.
+        :param float dtime:
+            Floating-point number of seconds since the last call to this,
+            function, for motion interpolation etc.
+        :param float x:
+            Document X position of the end-point of this stroke.
+        :param float y:
+            Document Y position of the end-point of this stroke.
+        :param float pressure:
+            Pressure, ranging from 0.0 to 1.0.
+        :param float xtilt:
+            X-axis tilt, ranging from -1.0 to 1.0.
+        :param float ytilt:
+            Y-axis tilt, ranging from -1.0 to 1.0.
+
+        """
         if not self.stroke:
             self.stroke = stroke.Stroke()
             self.stroke.start_recording(self.brush)
@@ -181,6 +254,7 @@ class Document():
         if split:
             self.split_stroke()
 
+
     def redo_last_stroke_with_different_brush(self, brush):
         cmd = self.get_last_command()
         if not isinstance(cmd, command.Stroke):
@@ -192,34 +266,34 @@ class Document():
         new_stroke.render(self.layer._surface)
         self.do(command.Stroke(self, new_stroke, snapshot_before))
 
-    def straight_line(self, src, dst):
-        self.split_stroke()
-        self.brush.reset() # reset dynamic states (eg. filtered velocity)
-
-        duration = 3.0
-        pressure = 0.3
-        N = 1000
-        x = numpy.linspace(src[0], dst[0], N)
-        y = numpy.linspace(src[1], dst[1], N)
-        # rest the brush in src for a minute, to avoid interpolation
-        # from the upper left corner (states are zero) (FIXME: the
-        # brush should handle this on its own, maybe?)
-        self.stroke_to(60.0, x[0], y[0], 0.0, 0.0, 0.0)
-        for i in xrange(N):
-            self.stroke_to(duration/N, x[i], y[i], pressure, 0.0, 0.0)
-        self.split_stroke()
-        self.brush.reset()
-
 
     def layer_modified_cb(self, *args):
+        """Forwards region modify notifications (area invalidations)
+
+        GUI code can respond to these notifications by appending callbacks to
+        `self.canvas_observers`. Each callback is invoked with the bounding box
+        of the changed region: ``cb(x, y, w, h)``, or ``cb(0, 0, 0, 0)`` to
+        denote that everything needs to be redrawn.
+
+        See also: `invalidate_all()`.
+
+        """
         # for now, any layer modification is assumed to be visible
         for f in self.canvas_observers:
             f(*args)
 
+
     def invalidate_all(self):
+        """Marks everything as invalid.
+
+        Invokes the callbacks in `self.canvas_observers` passing the notation
+        for "everything" as arguments. See `layer_modified_cb()` for details.
+
+        """
         for f in self.canvas_observers:
             f(0, 0, 0, 0)
 
+
     def undo(self):
         self.split_stroke()
         while 1:
@@ -238,11 +312,24 @@ class Document():
         self.split_stroke()
         self.command_stack.do(cmd)
 
+
+    def update_last_command(self, **kwargs):
+        self.split_stroke()
+        return self.command_stack.update_last_command(**kwargs)
+
+
     def get_last_command(self):
         self.split_stroke()
         return self.command_stack.get_last_command()
 
+
     def get_bbox(self):
+        """Returns the dynamic bounding box of the document.
+
+        This is currently the union of all the bounding boxes of all of the
+        layers. It disregards the user-chosen frame.
+
+        """
         res = helpers.Rect()
         for layer in self.layers:
             # OPTIMIZE: only visible layers...
@@ -251,40 +338,58 @@ class Document():
             res.expandToIncludeRect(bbox)
         return res
 
+
     def get_effective_bbox(self):
         """Return the effective bounding box of the document.
+
         If the frame is enabled, this is the bounding box of the frame, 
-        else the (dynamic) bounding box of the document."""
+        else the (dynamic) bounding box of the document.
+
+        """
         return self.get_frame() if self.frame_enabled else self.get_bbox()
 
-    def blit_tile_into(self, dst_8bit, tx, ty, mipmap_level=0, layers=None, background=None):
+    def blit_tile_into(self, dst, dst_has_alpha, tx, ty, mipmap_level=0, layers=None, background=None):
+        assert dst_has_alpha is False
         if layers is None:
             layers = self.layers
         if background is None:
             background = self.background
 
-        assert dst_8bit.dtype == 'uint8'
-        dst = numpy.empty((N, N, 3), dtype='uint16')
+        assert dst.shape[-1] == 4
+        if dst.dtype == 'uint8':
+            dst_8bit = dst
+            dst = numpy.empty((N, N, 4), dtype='uint16')
+        else:
+            dst_8bit = None
 
-        background.blit_tile_into(dst, tx, ty, mipmap_level)
+        background.blit_tile_into(dst, dst_has_alpha, tx, ty, mipmap_level)
 
         for layer in layers:
-            surface = layer._surface
-            surface.composite_tile(dst, tx, ty,
-                    mipmap_level=mipmap_level,
-                    opacity=layer.effective_opacity,
-                    mode=layer.compositeop)
+            layer.composite_tile(dst, dst_has_alpha, tx, ty, mipmap_level)
+
+        if dst_8bit is not None:
+            mypaintlib.tile_convert_rgbu16_to_rgbu8(dst, dst_8bit)
+
+    def get_rendered_image_behind_current_layer(self, tx, ty):
+        dst = numpy.empty((N, N, 4), dtype='uint16')
+        l = self.layers[0:self.layer_idx]
+        self.blit_tile_into(dst, False, tx, ty, layers=l)
+        return dst
 
-        mypaintlib.tile_convert_rgb16_to_rgb8(dst, dst_8bit)
 
     def add_layer(self, insert_idx=None, after=None, name=''):
         self.do(command.AddLayer(self, insert_idx, after, name))
 
+
     def remove_layer(self,layer=None):
-        if len(self.layers) > 1:
-            self.do(command.RemoveLayer(self,layer))
-        else:
-            self.clear_layer()
+        self.do(command.RemoveLayer(self, layer))
+
+
+    def rename_layer(self, layer, name):
+        self.do(command.RenameLayer(self, name, layer))
+
+    def convert_layer_to_normal_mode(self):
+        self.do(command.ConvertLayerToNormalMode(self, self.layer))
 
     def merge_layer_down(self):
         dst_idx = self.layer_idx - 1
@@ -293,38 +398,58 @@ class Document():
         self.do(command.MergeLayer(self, dst_idx))
         return True
 
+
     def load_layer_from_pixbuf(self, pixbuf, x=0, y=0):
         arr = helpers.gdkpixbuf2numpy(pixbuf)
         s = tiledsurface.Surface()
-        s.load_from_numpy(arr, x, y)
+        bbox = s.load_from_numpy(arr, x, y)
         self.do(command.LoadLayer(self, s))
+        return bbox
+
 
     def load_layer_from_png(self, filename, x=0, y=0, feedback_cb=None):
         s = tiledsurface.Surface()
-        s.load_from_png(filename, x, y, feedback_cb)
+        bbox = s.load_from_png(filename, x, y, feedback_cb)
         self.do(command.LoadLayer(self, s))
+        return bbox
+
 
     def set_layer_visibility(self, visible, layer):
+        """Sets the visibility of a layer."""
         cmd = self.get_last_command()
         if isinstance(cmd, command.SetLayerVisibility) and cmd.layer is layer:
-            self.undo()
-        self.do(command.SetLayerVisibility(self, visible, layer))
+            self.update_last_command(visible=visible)
+        else:
+            self.do(command.SetLayerVisibility(self, visible, layer))
+
 
     def set_layer_locked(self, locked, layer):
+        """Sets the input-locked status of a layer."""
         cmd = self.get_last_command()
         if isinstance(cmd, command.SetLayerLocked) and cmd.layer is layer:
-            self.undo()
-        self.do(command.SetLayerLocked(self, locked, layer))
+            self.update_last_command(locked=locked)
+        else:
+            self.do(command.SetLayerLocked(self, locked, layer))
+
 
     def set_layer_opacity(self, opacity, layer=None):
-        """Sets the opacity of a layer. If layer=None, works on the current layer"""
+        """Sets the opacity of a layer.
+
+        If layer=None, works on the current layer.
+
+        """
         cmd = self.get_last_command()
         if isinstance(cmd, command.SetLayerOpacity):
             self.undo()
         self.do(command.SetLayerOpacity(self, opacity, layer))
 
+
     def set_layer_compositeop(self, compositeop, layer=None):
-        """Sets the composition-operation of a layer. If layer=None, works on the current layer"""
+        """Sets the compositing operator for a layer.
+
+        If layer=None, works on the current layer.
+
+        """
         if compositeop not in VALID_COMPOSITE_OPS:
             compositeop = DEFAULT_COMPOSITE_OP
         cmd = self.get_last_command()
@@ -332,23 +457,27 @@ class Document():
             self.undo()
         self.do(command.SetLayerCompositeOp(self, compositeop, layer))
 
-    def set_background(self, obj):
+
+    def set_background(self, obj, make_default=False):
         # This is not an undoable action. One reason is that dragging
         # on the color chooser would get tons of undo steps.
-
         if not isinstance(obj, backgroundsurface.Background):
             obj = backgroundsurface.Background(obj)
         self.background = obj
-
+        if make_default:
+            self.default_background = obj
         self.invalidate_all()
 
+
     def load_from_pixbuf(self, pixbuf):
         """Load a document from a pixbuf."""
         self.clear()
-        self.load_layer_from_pixbuf(pixbuf)
-        self.set_frame(*self.get_bbox())
+        bbox = self.load_layer_from_pixbuf(pixbuf)
+        self.set_frame(*bbox)
+
 
     def is_layered(self):
+        """True if there are more than one nonempty layers."""
         count = 0
         for l in self.layers:
             if not l.is_empty():
@@ -356,13 +485,26 @@ class Document():
         return count > 1
 
     def is_empty(self):
+        """True if there is only one layer and it is empty."""
         return len(self.layers) == 1 and self.layer.is_empty()
 
     def save(self, filename, **kwargs):
+        """Save the document to a file.
+
+        :param str filename:
+            The filename to save to. The extension is used to determine format,
+            and a ``save_*()`` method is chosen to perform the save.
+        :param dict kwargs:
+            Passed on to the chosen save method.
+        :raise SaveLoadError:
+            The error string will be set to something descriptive and
+            presentable to the user.
+
+        """
         self.split_stroke()
         junk, ext = os.path.splitext(filename)
         ext = ext.lower().replace('.', '')
-        save = getattr(self, 'save_' + ext, self.unsupported)
+        save = getattr(self, 'save_' + ext, self._unsupported)
         try:
             save(filename, **kwargs)
         except gobject.GError, e:
@@ -377,14 +519,27 @@ class Document():
             raise SaveLoadError, _('Unable to save: %s') % e.strerror
         self.unsaved_painting_time = 0.0
 
+
     def load(self, filename, **kwargs):
+        """Load the document from a file.
+
+        :param str filename:
+            The filename to load from. The extension is used to determine
+            format, and a ``load_*()`` method is chosen to perform the load.
+        :param dict kwargs:
+            Passed on to the chosen loader method.
+        :raise SaveLoadError:
+            The error string will be set to something descriptive and
+            presentable to the user.
+
+        """
         if not os.path.isfile(filename):
             raise SaveLoadError, _('File does not exist: %s') % repr(filename)
         if not os.access(filename,os.R_OK):
             raise SaveLoadError, _('You do not have the necessary permissions to open file: %s') % repr(filename)
         junk, ext = os.path.splitext(filename)
         ext = ext.lower().replace('.', '')
-        load = getattr(self, 'load_' + ext, self.unsupported)
+        load = getattr(self, 'load_' + ext, self._unsupported)
         try:
             load(filename, **kwargs)
         except gobject.GError, e:
@@ -397,7 +552,8 @@ class Document():
         self.unsaved_painting_time = 0.0
         self.call_doc_observers()
 
-    def unsupported(self, filename, *args, **kwargs):
+
+    def _unsupported(self, filename, *args, **kwargs):
         raise SaveLoadError, _('Unknown file format extension: %s') % repr(filename)
 
     def render_as_pixbuf(self, *args, **kwargs):
@@ -446,12 +602,15 @@ class Document():
 
     def load_png(self, filename, feedback_cb=None):
         self.clear()
-        self.load_layer_from_png(filename, 0, 0, feedback_cb)
-        self.set_frame(*self.get_bbox())
+        bbox = self.load_layer_from_png(filename, 0, 0, feedback_cb)
+        self.set_frame(*bbox)
 
     @staticmethod
     def _pixbuf_from_stream(fp, feedback_cb=None):
-        loader = gdk.PixbufLoader()
+        if gui.pygtkcompat.USE_GTK3:
+            loader = GdkPixbuf.PixbufLoader()
+        else:
+            loader = gdk.PixbufLoader()
         while True:
             if feedback_cb is not None:
                 feedback_cb()
@@ -476,14 +635,17 @@ class Document():
         if w == 0 or h == 0:
             x, y, w, h = 0, 0, N, N # allow to save empty documents
         pixbuf = self.render_as_pixbuf(x, y, w, h, **kwargs)
-        pixbuf.save(filename, 'jpeg', options={'quality':str(quality)})
+        options = {"quality": str(quality)}
+        gui.pygtkcompat.gdk.pixbuf.save(pixbuf, filename, 'jpeg', **options)
 
     save_jpeg = save_jpg
 
     def save_ora(self, filename, options=None, **kwargs):
         print 'save_ora:'
         t0 = time.time()
-        tempdir = tempfile.mkdtemp(u'mypaint')
+        tempdir = tempfile.mkdtemp('mypaint')
+        if not isinstance(tempdir, unicode):
+            tempdir = tempdir.decode(sys.getfilesystemencoding())
         # use .tmp extension, so we don't overwrite a valid file if there is an exception
         z = zipfile.ZipFile(filename + '.tmpsave', 'w', compression=zipfile.ZIP_STORED)
         # work around a permission bug in the zipfile library: http://bugs.python.org/issue3394
@@ -502,7 +664,7 @@ class Document():
         def store_pixbuf(pixbuf, name):
             tmp = join(tempdir, 'tmp.png')
             t1 = time.time()
-            pixbuf.save(tmp, 'png')
+            gui.pygtkcompat.gdk.pixbuf.save(pixbuf, tmp, 'png')
             print '  %.3fs pixbuf saving %s' % (time.time() - t1, name)
             z.write(tmp, name)
             os.remove(tmp)
@@ -515,7 +677,9 @@ class Document():
             z.write(tmp, name)
             os.remove(tmp)
 
-        def add_layer(x, y, opac, surface, name, layer_name, visible=True, compositeop=DEFAULT_COMPOSITE_OP, rect=[]):
+        def add_layer(x, y, opac, surface, name, layer_name, visible=True,
+                      locked=False, selected=False,
+                      compositeop=DEFAULT_COMPOSITE_OP, rect=[]):
             layer = ET.Element('layer')
             stack.append(layer)
             store_surface(surface, name, rect)
@@ -533,6 +697,10 @@ class Document():
                 a['visibility'] = 'visible'
             else:
                 a['visibility'] = 'hidden'
+            if locked:
+                a['edit-locked'] = 'true'
+            if selected:
+                a['selected'] = 'true'
             return layer
 
         for idx, l in enumerate(reversed(self.layers)):
@@ -540,7 +708,11 @@ class Document():
                 continue
             opac = l.opacity
             x, y, w, h = l.get_bbox()
-            el = add_layer(x-x0, y-y0, opac, l._surface, 'data/layer%03d.png' % idx, l.name, l.visible, l.compositeop, rect=(x, y, w, h))
+            sel = (idx == self.layer_idx)
+            el = add_layer(x-x0, y-y0, opac, l._surface,
+                           'data/layer%03d.png' % idx, l.name, l.visible,
+                           locked=l.locked, selected=sel,
+                           compositeop=l.compositeop, rect=(x, y, w, h))
             # strokemap
             sio = StringIO()
             l.save_strokemap_to_file(sio, -x, -y)
@@ -554,7 +726,9 @@ class Document():
         # save as fully rendered layer
         x, y, w, h = self.get_bbox()
         l = add_layer(x-x0, y-y0, 1.0, bg, 'data/background.png', 'background',
-                      DEFAULT_COMPOSITE_OP, rect=(x,y,w,h))
+                      locked=True, selected=False,
+                      compositeop=DEFAULT_COMPOSITE_OP,
+                      rect=(x,y,w,h))
         x, y, w, h = bg.get_pattern_bbox()
         # save as single pattern (with corrected origin)
         store_surface(bg, 'data/background_tile.png', rect=(x+x0, y+y0, w, h))
@@ -582,11 +756,19 @@ class Document():
 
         return thumbnail_pixbuf
 
+    @staticmethod
+    def __xsd2bool(v):
+        v = str(v).lower()
+        if v in ['true', '1']: return True
+        else: return False
+
     def load_ora(self, filename, feedback_cb=None):
         """Loads from an OpenRaster file"""
         print 'load_ora:'
         t0 = time.time()
-        tempdir = tempfile.mkdtemp(u'mypaint')
+        tempdir = tempfile.mkdtemp('mypaint')
+        if not isinstance(tempdir, unicode):
+            tempdir = tempdir.decode(sys.getfilesystemencoding())
         z = zipfile.ZipFile(filename)
         print 'mimetype:', z.read('mimetype').strip()
         xml = z.read('stack.xml')
@@ -596,14 +778,6 @@ class Document():
         w = int(image.attrib['w'])
         h = int(image.attrib['h'])
 
-        def round_up_to_n(value, n):
-            assert value >= 0, "function undefined for negative numbers"
-
-            residual = value % n
-            if residual:
-                value = value - residual + n
-            return int(value)
-
         def get_pixbuf(filename):
             t1 = time.time()
 
@@ -638,9 +812,9 @@ class Document():
 
         self.clear() # this leaves one empty layer
         no_background = True
-        # FIXME: don't require tile alignment for frame
-        self.set_frame(width=round_up_to_n(w, N), height=round_up_to_n(h, N))
+        self.set_frame(width=w, height=h)
 
+        selected_layer = None
         for layer in get_layers_list(stack):
             a = layer.attrib
 
@@ -665,6 +839,8 @@ class Document():
             compositeop = str(a.get('composite-op', DEFAULT_COMPOSITE_OP))
             if compositeop not in VALID_COMPOSITE_OPS:
                 compositeop = DEFAULT_COMPOSITE_OP
+            selected = self.__xsd2bool(a.get("selected", 'false'))
+            locked = self.__xsd2bool(a.get("edit-locked", 'false'))
 
             visible = not 'hidden' in a.get('visibility', 'visible')
             self.add_layer(insert_idx=0, name=name)
@@ -682,6 +858,9 @@ class Document():
             self.set_layer_opacity(helpers.clamp(opac, 0.0, 1.0), layer)
             self.set_layer_compositeop(compositeop, layer)
             self.set_layer_visibility(visible, layer)
+            self.set_layer_locked(locked, layer)
+            if selected:
+                selected_layer = layer
             print '  %.3fs loading and converting layer png' % (time.time() - t1)
             # strokemap
             fname = a.get('mypaint_strokemap_v2', None)
@@ -703,6 +882,12 @@ class Document():
             self.remove_layer()
             # this leaves the topmost layer selected
 
+        if selected_layer is not None:
+            for i, layer in zip(range(len(self.layers)), self.layers):
+                if layer is selected_layer:
+                    self.select_layer(i)
+                    break
+
         z.close()
 
         # remove empty directories created by zipfile's extract()
