@@ -1,3 +1,11 @@
+
+#ifndef HELPERS2_HPP
+#define HELPERS2_HPP
+
+// Making the helpers of brushlib also bound by Python
+#include "helpers.c"
+#include <glib.h>
+
 // Special HSV -> RGB converter for use with the color selector classes
 // Takes values in the range [ 0.0 , 1.0 ]
 // Gives values in the range [ 0.0 , 255.0 ]
@@ -69,3 +77,20 @@ void hsv_to_rgb_range_one(float *h_, float *s_, float *v_)
   *v_ = b*255.0f;
 }
 
+typedef struct { int x, y, w, h; } Rect;
+// originally from my mass project (mass.sourceforge.net)
+void ExpandRectToIncludePoint(Rect * r, int x, int y)
+{
+  if (r->w == 0) {
+    r->w = 1; r->h = 1;
+    r->x = x; r->y = y;
+  } else {
+    if (x < r->x) { r->w += r->x-x; r->x = x; } else
+    if (x >= r->x+r->w) { r->w = x - r->x + 1; }
+
+    if (y < r->y) { r->h += r->y-y; r->y = y; } else
+    if (y >= r->y+r->h) { r->h = y - r->y + 1; }
+  }
+}
+
+#endif //HELPERS2_HPP
