@@ -1,9 +1,16 @@
 #include "Python.h"
 #include "numpy/arrayobject.h"
-#include "../brushlib/brushlib.hpp"
+#include "mapping.hpp"
+#include "surface.hpp"
+#include "brush.hpp"
 #include "python_brush.hpp"
-#include "brushmodes.hpp"
+#include "helpers2.hpp"
 #include "tiledsurface.hpp"
+
+#ifdef HAVE_GEGL
+#include "geglbackedsurface.hpp"
+#endif
+
 #include "pixops.hpp"
 #include "colorring.hpp"
 #include "colorchanger_wash.hpp"
