@@ -7,46 +7,24 @@
  * (at your option) any later version.
  */
 
-// downscale a tile to half its size using bilinear interpolation
-// used mainly for generating background mipmaps
-void tile_downscale_rgb16(PyObject *src, PyObject *dst, int dst_x, int dst_y) {
+// make the "heavy_debug" readable from python
 #ifdef HEAVY_DEBUG
-  assert(PyArray_DIM(src, 0) == TILE_SIZE);
-  assert(PyArray_DIM(src, 1) == TILE_SIZE);
-  assert(PyArray_TYPE(src) == NPY_UINT16);
-  assert(PyArray_ISCARRAY(src));
-
-  assert(PyArray_TYPE(dst) == NPY_UINT16);
-  assert(PyArray_ISCARRAY(dst));
+const bool heavy_debug = true;
+#else
+const bool heavy_debug = false;
 #endif
 
-  PyArrayObject* src_arr = ((PyArrayObject*)src);
-  PyArrayObject* dst_arr = ((PyArrayObject*)dst);
-
-  for (int y=0; y<TILE_SIZE/2; y++) {
-    uint16_t * src_p = (uint16_t*)(src_arr->data + (2*y)*src_arr->strides[0]);
-    uint16_t * dst_p = (uint16_t*)(dst_arr->data + (y+dst_y)*dst_arr->strides[0]);
-    dst_p += 3*dst_x;
-    for(int x=0; x<TILE_SIZE/2; x++) {
-      dst_p[0] = src_p[0]/4 + (src_p+3)[0]/4 + (src_p+3*TILE_SIZE)[0]/4 + (src_p+3*TILE_SIZE+3)[0]/4;
-      dst_p[1] = src_p[1]/4 + (src_p+3)[1]/4 + (src_p+3*TILE_SIZE)[1]/4 + (src_p+3*TILE_SIZE+3)[1]/4;
-      dst_p[2] = src_p[2]/4 + (src_p+3)[2]/4 + (src_p+3*TILE_SIZE)[2]/4 + (src_p+3*TILE_SIZE+3)[2]/4;
-      src_p += 6;
-      dst_p += 3;
-    }
-  }
-}
 // downscale a tile to half its size using bilinear interpolation
-// used mainly for generating tiledsurface mipmaps
+// used for generating mipmaps for tiledsurface and background
 void tile_downscale_rgba16(PyObject *src, PyObject *dst, int dst_x, int dst_y) {
 #ifdef HEAVY_DEBUG
-  assert(PyArray_DIM(src, 0) == TILE_SIZE);
-  assert(PyArray_DIM(src, 1) == TILE_SIZE);
+  assert(PyArray_DIM(src, 0) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(src, 1) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(src, 2) == 4);
   assert(PyArray_TYPE(src) == NPY_UINT16);
   assert(PyArray_ISCARRAY(src));
 
-  assert(PyArray_DIM(dst, 0) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 1) == TILE_SIZE);
+  assert(PyArray_DIM(dst, 2) == 4);
   assert(PyArray_TYPE(dst) == NPY_UINT16);
   assert(PyArray_ISCARRAY(dst));
 #endif
@@ -54,307 +32,279 @@ void tile_downscale_rgba16(PyObject *src, PyObject *dst, int dst_x, int dst_y) {
   PyArrayObject* src_arr = ((PyArrayObject*)src);
   PyArrayObject* dst_arr = ((PyArrayObject*)dst);
 
-  for (int y=0; y<TILE_SIZE/2; y++) {
+  for (int y=0; y<MYPAINT_TILE_SIZE/2; y++) {
     uint16_t * src_p = (uint16_t*)(src_arr->data + (2*y)*src_arr->strides[0]);
     uint16_t * dst_p = (uint16_t*)(dst_arr->data + (y+dst_y)*dst_arr->strides[0]);
     dst_p += 4*dst_x;
-    for(int x=0; x<TILE_SIZE/2; x++) {
-      dst_p[0] = src_p[0]/4 + (src_p+4)[0]/4 + (src_p+4*TILE_SIZE)[0]/4 + (src_p+4*TILE_SIZE+4)[0]/4;
-      dst_p[1] = src_p[1]/4 + (src_p+4)[1]/4 + (src_p+4*TILE_SIZE)[1]/4 + (src_p+4*TILE_SIZE+4)[1]/4;
-      dst_p[2] = src_p[2]/4 + (src_p+4)[2]/4 + (src_p+4*TILE_SIZE)[2]/4 + (src_p+4*TILE_SIZE+4)[2]/4;
-      dst_p[3] = src_p[3]/4 + (src_p+4)[3]/4 + (src_p+4*TILE_SIZE)[3]/4 + (src_p+4*TILE_SIZE+4)[3]/4;
+    for(int x=0; x<MYPAINT_TILE_SIZE/2; x++) {
+      dst_p[0] = src_p[0]/4 + (src_p+4)[0]/4 + (src_p+4*MYPAINT_TILE_SIZE)[0]/4 + (src_p+4*MYPAINT_TILE_SIZE+4)[0]/4;
+      dst_p[1] = src_p[1]/4 + (src_p+4)[1]/4 + (src_p+4*MYPAINT_TILE_SIZE)[1]/4 + (src_p+4*MYPAINT_TILE_SIZE+4)[1]/4;
+      dst_p[2] = src_p[2]/4 + (src_p+4)[2]/4 + (src_p+4*MYPAINT_TILE_SIZE)[2]/4 + (src_p+4*MYPAINT_TILE_SIZE+4)[2]/4;
+      dst_p[3] = src_p[3]/4 + (src_p+4)[3]/4 + (src_p+4*MYPAINT_TILE_SIZE)[3]/4 + (src_p+4*MYPAINT_TILE_SIZE+4)[3]/4;
       src_p += 8;
       dst_p += 4;
     }
   }
 }
 
-void tile_composite_rgba16_over_rgb16(PyObject * src, PyObject * dst, float alpha) {
+
+#include "compositing.hpp"
+#include "blendmodes.hpp"
+
+
+
+// Composite one tile over another.
+
+template <typename B>
+static inline void
+tile_composite_data (PyObject *src,
+                     PyObject *dst,
+                     const bool dst_has_alpha,
+                     const float src_opacity)
+{
 #ifdef HEAVY_DEBUG
-  assert(PyArray_DIM(src, 0) == TILE_SIZE);
-  assert(PyArray_DIM(src, 1) == TILE_SIZE);
+  assert(PyArray_DIM(src, 0) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(src, 1) == MYPAINT_TILE_SIZE);
   assert(PyArray_DIM(src, 2) == 4);
   assert(PyArray_TYPE(src) == NPY_UINT16);
   assert(PyArray_ISCARRAY(src));
 
-  assert(PyArray_DIM(dst, 0) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 1) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 2) == 3);
+  assert(PyArray_DIM(dst, 0) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(dst, 1) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(dst, 2) == 4);
   assert(PyArray_TYPE(dst) == NPY_UINT16);
-  assert(PyArray_ISBEHAVED(dst));
-#endif
-  
+  assert(PyArray_ISCARRAY(dst));
+
   PyArrayObject* dst_arr = ((PyArrayObject*)dst);
-#ifdef HEAVY_DEBUG
-  assert(dst_arr->strides[1] == 3*sizeof(uint16_t));
-  assert(dst_arr->strides[2] ==   sizeof(uint16_t));
+  assert(dst_arr->strides[0] == 4*sizeof(fix15_short_t)*MYPAINT_TILE_SIZE);
+  assert(dst_arr->strides[1] == 4*sizeof(fix15_short_t));
+  assert(dst_arr->strides[2] ==   sizeof(fix15_short_t));
 #endif
 
-  uint32_t opac  = alpha * (1<<15) + 0.5;
-  opac = CLAMP(opac, 0, 1<<15);
-  if (opac == 0) return;
+  const fix15_short_t opac = fix15_short_clamp(src_opacity * fix15_one);
+  if (opac == 0)
+    return;
 
-  uint16_t * src_p  = (uint16_t*)((PyArrayObject*)src)->data;
-  char * p = dst_arr->data;
-  for (int y=0; y<TILE_SIZE; y++) {
-    uint16_t  * dst_p  = (uint16_t*) (p);
-    for (int x=0; x<TILE_SIZE; x++) {
-      // resultAlpha = 1.0 (thus it does not matter if resultColor is premultiplied alpha or not)
-      // resultColor = topColor + (1.0 - topAlpha) * bottomColor
-      const uint16_t srcAlpha = CLAMP((uint32_t)(opac * src_p[3]) / (1<<15), 0, 1<<15);
-      const uint32_t one_minus_topAlpha = (1<<15) - srcAlpha;
-      dst_p[0] = ((uint32_t)src_p[0]*opac + one_minus_topAlpha*dst_p[0]) / (1<<15);
-      dst_p[1] = ((uint32_t)src_p[1]*opac + one_minus_topAlpha*dst_p[1]) / (1<<15);
-      dst_p[2] = ((uint32_t)src_p[2]*opac + one_minus_topAlpha*dst_p[2]) / (1<<15);
-      src_p += 4;
-      dst_p += 3;
-    }
-    p += dst_arr->strides[0];
+  const fix15_short_t* const src_p = (fix15_short_t *)
+                                        ((PyArrayObject*)src)->data;
+  fix15_short_t*       const dst_p = (fix15_short_t *)
+                                        ((PyArrayObject*)dst)->data;
+  if (dst_has_alpha) {
+    BufferComp<BufferCompOutputRGBA, MYPAINT_TILE_SIZE*MYPAINT_TILE_SIZE*4, B>
+        ::composite_src_over(src_p, dst_p, opac);
+  }
+  else {
+    BufferComp<BufferCompOutputRGBX, MYPAINT_TILE_SIZE*MYPAINT_TILE_SIZE*4, B>
+        ::composite_src_over(src_p, dst_p, opac);
   }
 }
 
-void tile_composite_rgba16_multiply_rgb16(PyObject * src, PyObject * dst, float alpha) {
-#ifdef HEAVY_DEBUG
-  assert(PyArray_DIM(src, 0) == TILE_SIZE);
-  assert(PyArray_DIM(src, 1) == TILE_SIZE);
-  assert(PyArray_DIM(src, 2) == 4);
-  assert(PyArray_TYPE(src) == NPY_UINT16);
-  assert(PyArray_ISCARRAY(src));
 
-  assert(PyArray_DIM(dst, 0) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 1) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 2) == 3);
-  assert(PyArray_TYPE(dst) == NPY_UINT16);
-  assert(PyArray_ISBEHAVED(dst));
-#endif
-  
-  PyArrayObject* dst_arr = ((PyArrayObject*)dst);
-#ifdef HEAVY_DEBUG
-  assert(dst_arr->strides[1] == 3*sizeof(uint16_t));
-  assert(dst_arr->strides[2] ==   sizeof(uint16_t));
-#endif
-
-  uint32_t opac  = alpha * (1<<15) + 0.5;
-  opac = CLAMP(opac, 0, 1<<15);
-  if (opac == 0) return;
-
-  uint16_t * src_p  = (uint16_t*)((PyArrayObject*)src)->data;
-  char * p = dst_arr->data;
-  for (int y=0; y<TILE_SIZE; y++) {
-    uint16_t  * dst_p  = (uint16_t*) (p);
-    for (int x=0; x<TILE_SIZE; x++) {
-      // resultAlpha = 1.0 (thus it does not matter if resultColor is premultiplied alpha or not)
-      // resultColor = topColor + (1.0 - topAlpha) * bottomColor
-      const uint32_t one_minus_topAlpha = (1<<15) - src_p[3]*opac/(1<<15);
-      const uint32_t src_col0 = ((uint32_t) src_p[0] * opac) >> 15;
-      const uint32_t src_col1 = ((uint32_t) src_p[1] * opac) >> 15;
-      const uint32_t src_col2 = ((uint32_t) src_p[2] * opac) >> 15;
-      dst_p[0] = ((uint32_t)src_col0*dst_p[0] + one_minus_topAlpha*dst_p[0]) / (1<<15);
-      dst_p[1] = ((uint32_t)src_col1*dst_p[1] + one_minus_topAlpha*dst_p[1]) / (1<<15);
-      dst_p[2] = ((uint32_t)src_col2*dst_p[2] + one_minus_topAlpha*dst_p[2]) / (1<<15);
-      src_p += 4;
-      dst_p += 3;
-    }
-    p += dst_arr->strides[0];
-  }
+void
+tile_composite_normal (PyObject *src,
+                       PyObject *dst,
+                       const bool dst_has_alpha,
+                       const float src_opacity)
+{
+    tile_composite_data<NormalBlendMode>(src, dst, dst_has_alpha, src_opacity);
 }
 
-void tile_composite_rgba16_screen_rgb16(PyObject * src, PyObject * dst, float alpha) {
-#ifdef HEAVY_DEBUG
-  assert(PyArray_DIM(src, 0) == TILE_SIZE);
-  assert(PyArray_DIM(src, 1) == TILE_SIZE);
-  assert(PyArray_DIM(src, 2) == 4);
-  assert(PyArray_TYPE(src) == NPY_UINT16);
-  assert(PyArray_ISCARRAY(src));
 
-  assert(PyArray_DIM(dst, 0) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 1) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 2) == 3);
-  assert(PyArray_TYPE(dst) == NPY_UINT16);
-  assert(PyArray_ISBEHAVED(dst));
-#endif
-  
-  PyArrayObject* dst_arr = ((PyArrayObject*)dst);
-#ifdef HEAVY_DEBUG
-  assert(dst_arr->strides[1] == 3*sizeof(uint16_t));
-  assert(dst_arr->strides[2] ==   sizeof(uint16_t));
-#endif
-
-  uint32_t opac  = alpha * (1<<15) + 0.5;
-  opac = CLAMP(opac, 0, 1<<15);
-  if (opac == 0) return;
-
-  uint16_t * src_p  = (uint16_t*)((PyArrayObject*)src)->data;
-  char * p = dst_arr->data;
-  for (int y=0; y<TILE_SIZE; y++) {
-    uint16_t  * dst_p  = (uint16_t*) (p);
-    for (int x=0; x<TILE_SIZE; x++) {
-      // resultAlpha = 1.0 (thus it does not matter if resultColor is premultiplied alpha or not)
-      // resultColor = topColor + (1.0 - topAlpha) * bottomColor
-      const uint32_t col0   = ((uint32_t)src_p[0]*opac) + (((uint32_t)dst_p[0]) << 15);
-      const uint32_t col1   = ((uint32_t)src_p[1]*opac) + (((uint32_t)dst_p[1]) << 15);
-      const uint32_t col2   = ((uint32_t)src_p[2]*opac) + (((uint32_t)dst_p[2]) << 15);
-      const uint32_t src_col0 = ((uint32_t)src_p[0] * opac) >> 15;
-      const uint32_t src_col1 = ((uint32_t)src_p[1] * opac) >> 15;
-      const uint32_t src_col2 = ((uint32_t)src_p[2] * opac) >> 15;
-      dst_p[0] = (col0 - ((uint32_t)src_col0*dst_p[0])) / (1<<15);
-      dst_p[1] = (col1 - ((uint32_t)src_col1*dst_p[1])) / (1<<15);
-      dst_p[2] = (col2 - ((uint32_t)src_col2*dst_p[2])) / (1<<15);
-      src_p += 4;
-      dst_p += 3;
-    }
-    p += dst_arr->strides[0];
-  }
+void
+tile_composite_multiply (PyObject *src,
+                         PyObject *dst,
+                         const bool dst_has_alpha,
+                         const float src_opacity)
+{
+    tile_composite_data<MultiplyBlendMode>(src,dst,dst_has_alpha,src_opacity);
 }
 
-void tile_composite_rgba16_dodge_rgb16(PyObject * src, PyObject * dst, float alpha) {
-#ifdef HEAVY_DEBUG
-  assert(PyArray_DIM(src, 0) == TILE_SIZE);
-  assert(PyArray_DIM(src, 1) == TILE_SIZE);
-  assert(PyArray_DIM(src, 2) == 4);
-  assert(PyArray_TYPE(src) == NPY_UINT16);
-  assert(PyArray_ISCARRAY(src));
 
-  assert(PyArray_DIM(dst, 0) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 1) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 2) == 3);
-  assert(PyArray_TYPE(dst) == NPY_UINT16);
-  assert(PyArray_ISBEHAVED(dst));
-#endif
-  
-  PyArrayObject* dst_arr = ((PyArrayObject*)dst);
-#ifdef HEAVY_DEBUG
-  assert(dst_arr->strides[1] == 3*sizeof(uint16_t));
-  assert(dst_arr->strides[2] ==   sizeof(uint16_t));
-#endif
-
-  uint32_t opac  = alpha * (1<<15) + 0.5;
-  opac = CLAMP(opac, 0, 1<<15);
-  if (opac == 0) return;
-
-  uint16_t * src_p  = (uint16_t*)((PyArrayObject*)src)->data;
-  char * p = dst_arr->data;
-  for (int y=0; y<TILE_SIZE; y++) {
-    uint16_t  * dst_p  = (uint16_t*) (p);
-    for (int x=0; x<TILE_SIZE; x++) {
-      // resultAlpha = 1.0 (thus it does not matter if resultColor is premultiplied alpha or not)
-      // resultColor = topColor + (1.0 - topAlpha) * bottomColor
-      const uint32_t topAlpha32 = CLAMP((uint32_t)src_p[3]*opac,0,1<<30);
-      const uint32_t topAlpha   = CLAMP((topAlpha32 >> 15), 0, 1<<15);
-      const uint32_t one_minus_topAlpha = (1<<15) - topAlpha;
-      for (int c=0; c < 3; c++) {
-        const uint32_t topAlpha_minus_src = topAlpha32 - (uint32_t)src_p[c]*opac;
-        if ((topAlpha_minus_src >> 15) == 0 && dst_p[c] == 0) {
-          dst_p[c] = 0;
-        } else if ((topAlpha_minus_src >> 15) == 0) {
-          dst_p[c] = CLAMP((topAlpha32 + (uint32_t)dst_p[c] * one_minus_topAlpha)>>15, 0, (1<<15));
-        } else {
-          const uint32_t dst_times_topAlpha = (uint32_t)dst_p[c]*topAlpha;
-          if (dst_times_topAlpha > topAlpha_minus_src)
-            dst_p[c] = CLAMP((topAlpha32 + (uint32_t)dst_p[c]*one_minus_topAlpha) >> 15, 0, 1<<15);
-          else
-            dst_p[c] = CLAMP((uint32_t)topAlpha * (dst_times_topAlpha >> 15)/ (topAlpha_minus_src >> 15) + ((uint32_t)dst_p[c]*one_minus_topAlpha >> 15), 0, 1<<15);
-        }
-      }
-      src_p += 4;
-      dst_p += 3;
-    }
-    p += dst_arr->strides[0];
-  }
+void
+tile_composite_screen (PyObject *src,
+                       PyObject *dst,
+                       const bool dst_has_alpha,
+                       const float src_opacity)
+{
+    tile_composite_data<ScreenBlendMode>(src,dst,dst_has_alpha,src_opacity);
 }
 
-void tile_composite_rgba16_burn_rgb16(PyObject * src, PyObject * dst, float alpha) {
-#ifdef HEAVY_DEBUG
-  assert(PyArray_DIM(src, 0) == TILE_SIZE);
-  assert(PyArray_DIM(src, 1) == TILE_SIZE);
-  assert(PyArray_DIM(src, 2) == 4);
-  assert(PyArray_TYPE(src) == NPY_UINT16);
-  assert(PyArray_ISCARRAY(src));
 
-  assert(PyArray_DIM(dst, 0) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 1) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 2) == 3);
-  assert(PyArray_TYPE(dst) == NPY_UINT16);
-  assert(PyArray_ISBEHAVED(dst));
-#endif
-  
-  PyArrayObject* dst_arr = ((PyArrayObject*)dst);
-#ifdef HEAVY_DEBUG
-  assert(dst_arr->strides[1] == 3*sizeof(uint16_t));
-  assert(dst_arr->strides[2] ==   sizeof(uint16_t));
-#endif
-
-  uint32_t opac  = alpha * (1<<15) + 0.5;
-  opac = CLAMP(opac, 0, 1<<15);
-  if (opac == 0) return;
-
-  uint16_t * src_p  = (uint16_t*)((PyArrayObject*)src)->data;
-  char * p = dst_arr->data;
-  for (int y=0; y<TILE_SIZE; y++) {
-    uint16_t  * dst_p  = (uint16_t*) (p);
-    for (int x=0; x<TILE_SIZE; x++) {
-      // resultAlpha = 1.0 (thus it does not matter if resultColor is premultiplied alpha or not)
-      // resultColor = topColor + (1.0 - topAlpha) * bottomColor
-      const uint32_t topAlpha32 = (uint32_t)src_p[3]*opac;
-      const uint32_t topAlpha   = topAlpha32 >> 15;
-      const uint32_t one_minus_topAlpha = (1<<15) - topAlpha;
-      for (int c=0; c<3; c++) {
-        const uint32_t src_col32 = (uint32_t)src_p[c] * opac;
-        const uint32_t src_col   = src_col32 >> 15;
-        if (src_col == 0 && dst_p[c] >= (1 << 15) - 1) {
-          dst_p[c] = 1<<15;
-        } else if (src_col == 0) {
-          dst_p[c] = (dst_p[c] * one_minus_topAlpha)/(1<<15);
-        } else {
-          const uint32_t one_minus_dstcol = (1<<15) - dst_p[c];
-          if (one_minus_dstcol * topAlpha > src_col32)
-            dst_p[c] = (one_minus_topAlpha * dst_p[c]) / (1<<15);
-          else {
-            const uint32_t min_value = one_minus_dstcol * topAlpha / src_col;
-            dst_p[c] = CLAMP((topAlpha32 - topAlpha * min_value + one_minus_topAlpha * dst_p[c]) / (1<<15), 0, 1<<15);
-          }
-        }
-      }
-      src_p += 4;
-      dst_p += 3;
-    }
-    p += dst_arr->strides[0];
-  }
+void
+tile_composite_overlay (PyObject *src,
+                        PyObject *dst,
+                        const bool dst_has_alpha,
+                        const float src_opacity)
+{
+    tile_composite_data<OverlayBlendMode>(src,dst,dst_has_alpha,src_opacity);
 }
 
-// used to copy the background before starting to composite over it
+
+void
+tile_composite_hard_light (PyObject *src,
+                           PyObject *dst,
+                           const bool dst_has_alpha,
+                           const float src_opacity)
+{
+    tile_composite_data<HardLightBlendMode>(src,dst,dst_has_alpha,src_opacity);
+}
+
+
+void
+tile_composite_lighten (PyObject *src,
+                        PyObject *dst,
+                        const bool dst_has_alpha,
+                        const float src_opacity)
+{
+    tile_composite_data<LightenBlendMode>
+        (src, dst, dst_has_alpha, src_opacity);
+}
+
+
+void
+tile_composite_darken (PyObject *src,
+                       PyObject *dst,
+                       const bool dst_has_alpha,
+                       const float src_opacity)
+{
+    tile_composite_data<DarkenBlendMode>
+        (src, dst, dst_has_alpha, src_opacity);
+}
+
+
+void
+tile_composite_soft_light (PyObject *src,
+                           PyObject *dst,
+                           const bool dst_has_alpha,
+                           const float src_opacity)
+{
+    tile_composite_data<SoftLightBlendMode>
+        (src, dst, dst_has_alpha, src_opacity);
+}
+
+
+void
+tile_composite_color_dodge (PyObject *src,
+                            PyObject *dst,
+                            const bool dst_has_alpha,
+                            const float src_opacity)
+{
+    tile_composite_data<ColorDodgeBlendMode>
+        (src, dst, dst_has_alpha, src_opacity);
+}
+
+
+void
+tile_composite_color_burn (PyObject *src,
+                           PyObject *dst,
+                           const bool dst_has_alpha,
+                           const float src_opacity)
+{
+    tile_composite_data<ColorBurnBlendMode>
+        (src, dst, dst_has_alpha, src_opacity);
+}
+
+
+void
+tile_composite_difference (PyObject *src,
+                           PyObject *dst,
+                           const bool dst_has_alpha,
+                           const float src_opacity)
+{
+    tile_composite_data<DifferenceBlendMode>
+        (src, dst, dst_has_alpha, src_opacity);
+}
+
+
+void
+tile_composite_exclusion (PyObject *src,
+                          PyObject *dst,
+                          const bool dst_has_alpha,
+                          const float src_opacity)
+{
+    tile_composite_data<ExclusionBlendMode>
+        (src, dst, dst_has_alpha, src_opacity);
+}
+
+void
+tile_composite_hue (PyObject *src,
+                    PyObject *dst,
+                    const bool dst_has_alpha,
+                    const float src_opacity)
+{
+    tile_composite_data<HueBlendMode>
+        (src, dst, dst_has_alpha, src_opacity);
+}
+
+
+void
+tile_composite_saturation (PyObject *src,
+                           PyObject *dst,
+                           const bool dst_has_alpha,
+                           const float src_opacity)
+{
+    tile_composite_data<SaturationBlendMode>
+        (src, dst, dst_has_alpha, src_opacity);
+}
+
+
+void
+tile_composite_color (PyObject *src,
+                      PyObject *dst,
+                      const bool dst_has_alpha,
+                      const float src_opacity)
+{
+    tile_composite_data<ColorBlendMode>
+        (src, dst, dst_has_alpha, src_opacity);
+}
+
+
+void
+tile_composite_luminosity (PyObject *src,
+                           PyObject *dst,
+                           const bool dst_has_alpha,
+                           const float src_opacity)
+{
+    tile_composite_data<LuminosityBlendMode>
+        (src, dst, dst_has_alpha, src_opacity);
+}
+
+
+
+
+// used to e.g. copy the background before starting to composite over it
 //
-// simply array copying (numpy assignment operator is about 13 times slower, sadly)
+// simply array copying (numpy assignment operator) is about 13 times slower, sadly
 // The above comment is true when the array is sliced; it's only about two
 // times faster now, in the current usecae.
-void tile_blit_rgb16_into_rgb16(PyObject * src, PyObject * dst) {
+void tile_copy_rgba16_into_rgba16(PyObject * src, PyObject * dst) {
   PyArrayObject* src_arr = ((PyArrayObject*)src);
   PyArrayObject* dst_arr = ((PyArrayObject*)dst);
 
 #ifdef HEAVY_DEBUG
-  assert(PyArray_DIM(dst, 0) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 1) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 2) == 3);
+  assert(PyArray_DIM(dst, 0) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(dst, 1) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(dst, 2) == 4);
   assert(PyArray_TYPE(dst) == NPY_UINT16);
   assert(PyArray_ISCARRAY(dst));
-  assert(dst_arr->strides[1] == 3*sizeof(uint16_t));
+  assert(dst_arr->strides[1] == 4*sizeof(uint16_t));
   assert(dst_arr->strides[2] ==   sizeof(uint16_t));
 
-  assert(PyArray_DIM(src, 0) == TILE_SIZE);
-  assert(PyArray_DIM(src, 1) == TILE_SIZE);
-  assert(PyArray_DIM(src, 2) == 3);
+  assert(PyArray_DIM(src, 0) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(src, 1) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(src, 2) == 4);
   assert(PyArray_TYPE(src) == NPY_UINT16);
   assert(PyArray_ISCARRAY(dst));
-  assert(src_arr->strides[1] == 3*sizeof(uint16_t));
+  assert(src_arr->strides[1] == 4*sizeof(uint16_t));
   assert(src_arr->strides[2] ==   sizeof(uint16_t));
 #endif
 
-  memcpy(dst_arr->data, src_arr->data, TILE_SIZE*TILE_SIZE*3*sizeof(uint16_t));
+  memcpy(dst_arr->data, src_arr->data, MYPAINT_TILE_SIZE*MYPAINT_TILE_SIZE*4*sizeof(uint16_t));
   /* the code below can be used if it is not ISCARRAY, but only ISBEHAVED:
   char * src_p = src_arr->data;
   char * dst_p = dst_arr->data;
-  for (int y=0; y<TILE_SIZE; y++) {
-    memcpy(dst_p, src_p, TILE_SIZE*3);
+  for (int y=0; y<MYPAINT_TILE_SIZE; y++) {
+    memcpy(dst_p, src_p, MYPAINT_TILE_SIZE*4);
     src_p += src_arr->strides[0];
     dst_p += dst_arr->strides[0];
   }
@@ -365,16 +315,16 @@ void tile_clear(PyObject * dst) {
   PyArrayObject* dst_arr = ((PyArrayObject*)dst);
 
 #ifdef HEAVY_DEBUG
-  assert(PyArray_DIM(dst, 0) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 1) == TILE_SIZE);
+  assert(PyArray_DIM(dst, 0) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(dst, 1) == MYPAINT_TILE_SIZE);
   assert(PyArray_TYPE(dst) == NPY_UINT8);
   assert(PyArray_ISBEHAVED(dst));
   assert(dst_arr->strides[1] <= 8);
 #endif
 
-  for (int y=0; y<TILE_SIZE; y++) {
+  for (int y=0; y<MYPAINT_TILE_SIZE; y++) {
     uint8_t  * dst_p = (uint8_t*)(dst_arr->data + y*dst_arr->strides[0]);
-    memset(dst_p, 0, TILE_SIZE*dst_arr->strides[1]);
+    memset(dst_p, 0, MYPAINT_TILE_SIZE*dst_arr->strides[1]);
     dst_p += dst_arr->strides[0];
   }
 }
@@ -406,16 +356,16 @@ void tile_convert_rgba16_to_rgba8(PyObject * src, PyObject * dst) {
   PyArrayObject* dst_arr = ((PyArrayObject*)dst);
 
 #ifdef HEAVY_DEBUG
-  assert(PyArray_DIM(dst, 0) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 1) == TILE_SIZE);
+  assert(PyArray_DIM(dst, 0) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(dst, 1) == MYPAINT_TILE_SIZE);
   assert(PyArray_DIM(dst, 2) == 4);
   assert(PyArray_TYPE(dst) == NPY_UINT8);
   assert(PyArray_ISBEHAVED(dst));
   assert(dst_arr->strides[1] == 4*sizeof(uint8_t));
   assert(dst_arr->strides[2] ==   sizeof(uint8_t));
 
-  assert(PyArray_DIM(src, 0) == TILE_SIZE);
-  assert(PyArray_DIM(src, 1) == TILE_SIZE);
+  assert(PyArray_DIM(src, 0) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(src, 1) == MYPAINT_TILE_SIZE);
   assert(PyArray_DIM(src, 2) == 4);
   assert(PyArray_TYPE(src) == NPY_UINT16);
   assert(PyArray_ISBEHAVED(src));
@@ -426,10 +376,10 @@ void tile_convert_rgba16_to_rgba8(PyObject * src, PyObject * dst) {
   precalculate_dithering_noise_if_required();
   int noise_idx = 0;
 
-  for (int y=0; y<TILE_SIZE; y++) {
+  for (int y=0; y<MYPAINT_TILE_SIZE; y++) {
     uint16_t * src_p = (uint16_t*)(src_arr->data + y*src_arr->strides[0]);
     uint8_t  * dst_p = (uint8_t*)(dst_arr->data + y*dst_arr->strides[0]);
-    for (int x=0; x<TILE_SIZE; x++) {
+    for (int x=0; x<MYPAINT_TILE_SIZE; x++) {
       uint32_t r, g, b, a;
       r = *src_p++;
       g = *src_p++;
@@ -511,82 +461,57 @@ void tile_convert_rgba16_to_rgba8(PyObject * src, PyObject * dst) {
 }
 
 // used after compositing (when displaying, or when saving solid PNG or JPG)
-void tile_convert_rgb16_to_rgb8(PyObject * src, PyObject * dst) {
+void tile_convert_rgbu16_to_rgbu8(PyObject * src, PyObject * dst) {
   PyArrayObject* src_arr = ((PyArrayObject*)src);
   PyArrayObject* dst_arr = ((PyArrayObject*)dst);
 
 #ifdef HEAVY_DEBUG
-  assert(PyArray_DIM(dst, 0) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 1) == TILE_SIZE);
+  assert(PyArray_DIM(dst, 0) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(dst, 1) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(dst, 2) == 4);
   assert(PyArray_TYPE(dst) == NPY_UINT8);
   assert(PyArray_ISBEHAVED(dst));
-  assert(PyArray_STRIDE(dst, 1) == PyArray_DIM(dst, 2)*sizeof(uint8_t));
+  assert(PyArray_STRIDE(dst, 1) == 4*sizeof(uint8_t));
   assert(PyArray_STRIDE(dst, 2) == sizeof(uint8_t));
 
-  assert(PyArray_DIM(src, 0) == TILE_SIZE);
-  assert(PyArray_DIM(src, 1) == TILE_SIZE);
-  assert(PyArray_DIM(src, 2) == 3);
+  assert(PyArray_DIM(src, 0) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(src, 1) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(src, 2) == 4);
   assert(PyArray_TYPE(src) == NPY_UINT16);
   assert(PyArray_ISBEHAVED(src));
-  assert(PyArray_STRIDE(src, 1) == 3*sizeof(uint16_t));
+  assert(PyArray_STRIDE(src, 1) == 4*sizeof(uint16_t));
   assert(PyArray_STRIDE(src, 2) ==   sizeof(uint16_t));
 #endif
 
   precalculate_dithering_noise_if_required();
   int noise_idx = 0;
 
-  bool dst_has_alpha = PyArray_DIM(dst, 2) == 4;
-
-  for (int y=0; y<TILE_SIZE; y++) {
+  for (int y=0; y<MYPAINT_TILE_SIZE; y++) {
     uint16_t * src_p = (uint16_t*)(src_arr->data + y*src_arr->strides[0]);
     uint8_t  * dst_p = (uint8_t*)(dst_arr->data + y*dst_arr->strides[0]);
-    if (dst_has_alpha) {
-      for (int x=0; x<TILE_SIZE; x++) {
-        uint32_t r, g, b;
-        r = *src_p++;
-        g = *src_p++;
-        b = *src_p++;
+    for (int x=0; x<MYPAINT_TILE_SIZE; x++) {
+      uint32_t r, g, b;
+      r = *src_p++;
+      g = *src_p++;
+      b = *src_p++;
+      src_p++; // alpha unused
 #ifdef HEAVY_DEBUG
-        assert(r<=(1<<15));
-        assert(g<=(1<<15));
-        assert(b<=(1<<15));
+      assert(r<=(1<<15));
+      assert(g<=(1<<15));
+      assert(b<=(1<<15));
 #endif
-          
-        /*
-        // rounding
-        const uint32_t add = (1<<15)/2;
-        */
-        // dithering
-        const uint32_t add = dithering_noise[noise_idx++];
-          
-        *dst_p++ = (r * 255 + add) / (1<<15);
-        *dst_p++ = (g * 255 + add) / (1<<15);
-        *dst_p++ = (b * 255 + add) / (1<<15);
-        *dst_p++ = 255;
-      }
-    } else {
-      for (int x=0; x<TILE_SIZE; x++) {
-        uint32_t r, g, b;
-        r = *src_p++;
-        g = *src_p++;
-        b = *src_p++;
-#ifdef HEAVY_DEBUG
-        assert(r<=(1<<15));
-        assert(g<=(1<<15));
-        assert(b<=(1<<15));
-#endif
-          
-        /*
-        // rounding
-        const uint32_t add = (1<<15)/2;
-        */
-        // dithering
-        const uint32_t add = dithering_noise[noise_idx++];
-          
-        *dst_p++ = (r * 255 + add) / (1<<15);
-        *dst_p++ = (g * 255 + add) / (1<<15);
-        *dst_p++ = (b * 255 + add) / (1<<15);
-      }
+      
+      /*
+      // rounding
+      const uint32_t add = (1<<15)/2;
+      */
+      // dithering
+      const uint32_t add = dithering_noise[noise_idx++];
+      
+      *dst_p++ = (r * 255 + add) / (1<<15);
+      *dst_p++ = (g * 255 + add) / (1<<15);
+      *dst_p++ = (b * 255 + add) / (1<<15);
+      *dst_p++ = 255;
     }
 #ifdef HEAVY_DEBUG
     assert(noise_idx <= dithering_noise_size);
@@ -602,16 +527,16 @@ void tile_convert_rgba8_to_rgba16(PyObject * src, PyObject * dst) {
   PyArrayObject* dst_arr = ((PyArrayObject*)dst);
 
 #ifdef HEAVY_DEBUG
-  assert(PyArray_DIM(dst, 0) == TILE_SIZE);
-  assert(PyArray_DIM(dst, 1) == TILE_SIZE);
+  assert(PyArray_DIM(dst, 0) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(dst, 1) == MYPAINT_TILE_SIZE);
   assert(PyArray_DIM(dst, 2) == 4);
   assert(PyArray_TYPE(dst) == NPY_UINT16);
   assert(PyArray_ISBEHAVED(dst));
   assert(dst_arr->strides[1] == 4*sizeof(uint16_t));
   assert(dst_arr->strides[2] ==   sizeof(uint16_t));
 
-  assert(PyArray_DIM(src, 0) == TILE_SIZE);
-  assert(PyArray_DIM(src, 1) == TILE_SIZE);
+  assert(PyArray_DIM(src, 0) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(src, 1) == MYPAINT_TILE_SIZE);
   assert(PyArray_DIM(src, 2) == 4);
   assert(PyArray_TYPE(src) == NPY_UINT8);
   assert(PyArray_ISBEHAVED(src));
@@ -619,10 +544,10 @@ void tile_convert_rgba8_to_rgba16(PyObject * src, PyObject * dst) {
   assert(src_arr->strides[2] ==   sizeof(uint8_t));
 #endif
 
-  for (int y=0; y<TILE_SIZE; y++) {
+  for (int y=0; y<MYPAINT_TILE_SIZE; y++) {
     uint8_t  * src_p = (uint8_t*)(src_arr->data + y*src_arr->strides[0]);
     uint16_t * dst_p = (uint16_t*)(dst_arr->data + y*dst_arr->strides[0]);
-    for (int x=0; x<TILE_SIZE; x++) {
+    for (int x=0; x<MYPAINT_TILE_SIZE; x++) {
       uint32_t r, g, b, a;
       r = *src_p++;
       g = *src_p++;
@@ -644,6 +569,110 @@ void tile_convert_rgba8_to_rgba16(PyObject * src, PyObject * dst) {
   }
 }
 
+// Flatten a premultiplied rgba layer, using "bg" as background.
+// (bg is assumed to be flat, bg.alpha is ignored)
+//
+// dst.color = dst OVER bg.color
+// dst.alpha = unmodified
+//
+void tile_rgba2flat(PyObject * dst, PyObject * bg) {
+#ifdef HEAVY_DEBUG
+  assert(PyArray_DIM(dst, 0) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(dst, 1) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(dst, 2) == 4);
+  assert(PyArray_TYPE(dst) == NPY_UINT16);
+  assert(PyArray_ISCARRAY(dst));
+
+  assert(PyArray_DIM(bg, 0) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(bg, 1) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(bg, 2) == 4);
+  assert(PyArray_TYPE(bg) == NPY_UINT16);
+  assert(PyArray_ISCARRAY(bg));
+#endif
+  
+  uint16_t * dst_p  = (uint16_t*)PyArray_DATA(dst);
+  uint16_t * bg_p  = (uint16_t*)PyArray_DATA(bg);
+  for (int i=0; i<MYPAINT_TILE_SIZE*MYPAINT_TILE_SIZE; i++) {
+    // resultAlpha = 1.0 (thus it does not matter if resultColor is premultiplied alpha or not)
+    // resultColor = topColor + (1.0 - topAlpha) * bottomColor
+    const uint32_t one_minus_top_alpha = (1<<15) - dst_p[3];
+    dst_p[0] += ((uint32_t)one_minus_top_alpha*bg_p[0]) / (1<<15);
+    dst_p[1] += ((uint32_t)one_minus_top_alpha*bg_p[1]) / (1<<15);
+    dst_p[2] += ((uint32_t)one_minus_top_alpha*bg_p[2]) / (1<<15);
+    dst_p += 4;
+    bg_p += 4;
+  }
+}
+
+// Make a flat layer translucent again. When calculating the new color
+// and alpha, it is assumed that the layer will be displayed OVER the
+// background "bg". Alpha is increased where required.
+//
+// dst.alpha = MIN(dst.alpha, minimum alpha required for correct result)
+// dst.color = calculated such that (dst_output OVER bg = dst_input.color)
+//
+void tile_flat2rgba(PyObject * dst, PyObject * bg) {
+#ifdef HEAVY_DEBUG
+  assert(PyArray_DIM(dst, 0) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(dst, 1) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(dst, 2) == 4);
+  assert(PyArray_TYPE(dst) == NPY_UINT16);
+  assert(PyArray_ISCARRAY(dst));
+
+  assert(PyArray_DIM(bg, 0) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(bg, 1) == MYPAINT_TILE_SIZE);
+  assert(PyArray_DIM(bg, 2) == 4);
+  assert(PyArray_TYPE(bg) == NPY_UINT16);
+  assert(PyArray_ISCARRAY(bg));
+#endif
+  
+  uint16_t * dst_p  = (uint16_t*)PyArray_DATA(dst);
+  uint16_t * bg_p  = (uint16_t*)PyArray_DATA(bg);
+  for (int i=0; i<MYPAINT_TILE_SIZE*MYPAINT_TILE_SIZE; i++) {
+
+    // 1. calculate final dst.alpha
+    uint16_t final_alpha = dst_p[3];
+    for (int i=0; i<3;i++) {
+      int32_t color_change = (int32_t)dst_p[i] - bg_p[i];
+      uint16_t minimal_alpha;
+      if (color_change > 0) {
+        minimal_alpha = (int64_t)color_change*(1<<15) / ((1<<15) - bg_p[i]);
+      } else if (color_change < 0) {
+        minimal_alpha = (int64_t)-color_change*(1<<15) / bg_p[i];
+      } else {
+        minimal_alpha = 0;
+      }
+      final_alpha = MAX(final_alpha, minimal_alpha);
+#ifdef HEAVY_DEBUG
+      assert(minimal_alpha <= (1<<15));
+      assert(final_alpha   <= (1<<15));
+#endif
+    }
+    
+    // 2. calculate dst.color and update dst
+    dst_p[3] = final_alpha;
+    if (final_alpha > 0) {
+      for (int i=0; i<3;i++) {
+        int32_t color_change = (int32_t)dst_p[i] - bg_p[i];
+        //int64_t res = bg_p[i] + (int64_t)color_change*(1<<15) / final_alpha;
+        // premultiplied with final_alpha
+        int64_t res = (uint32_t)bg_p[i]*final_alpha/(1<<15) + (int64_t)color_change;
+        res = CLAMP(res, 0, final_alpha); // fix rounding errors
+        dst_p[i] = res;
+#ifdef HEAVY_DEBUG
+        assert(dst_p[i] <= dst_p[3]);
+#endif
+      }
+    } else {
+      dst_p[0] = 0;
+      dst_p[1] = 0;
+      dst_p[2] = 0;
+    }
+    dst_p += 4;
+    bg_p += 4;
+  }
+}
+
 // used in strokemap.py
 //
 // Calculates a 1-bit bitmap of the stroke shape using two snapshots
@@ -672,8 +701,8 @@ void tile_perceptual_change_strokemap(PyObject * a, PyObject * b, PyObject * res
   uint16_t * b_p  = (uint16_t*)PyArray_DATA(b);
   uint8_t * res_p = (uint8_t*)PyArray_DATA(res);
 
-  for (int y=0; y<TILE_SIZE; y++) {
-    for (int x=0; x<TILE_SIZE; x++) {
+  for (int y=0; y<MYPAINT_TILE_SIZE; y++) {
+    for (int x=0; x<MYPAINT_TILE_SIZE; x++) {
 
       int32_t color_change = 0;
       // We want to compare a.color with b.color, but we only know
