@@ -9,14 +9,20 @@
 # This module implements an unbounded tiled surface for painting.
 
 from numpy import *
-import time, sys
+import time, sys, os
 import mypaintlib, helpers
+import math
 
 TILE_SIZE = N = mypaintlib.TILE_SIZE
-MAX_MIPMAP_LEVEL = mypaintlib.MAX_MIPMAP_LEVEL
+MAX_MIPMAP_LEVEL = 4
+
+use_gegl = True if os.environ.get('MYPAINT_ENABLE_GEGL', 0) else False
 
 from layer import DEFAULT_COMPOSITE_OP
-import pixbufsurface
+
+# Avoid pulling in PyGTK+ when using GI
+if not os.environ.get('MYPAINT_ENABLE_GEGL', 0):
+    import pixbufsurface
 
 
 class Tile:
@@ -34,123 +40,24 @@ class Tile:
         return Tile(copy_from=self)
 
 
-def composite_array_src_over(dst, src, mipmap_level=0, opacity=1.0):
-    """The default "svg:src-over" layer composite op implementation.
-    """
-    if dst.shape[2] == 4 and dst.dtype == 'uint16':
-        # rarely used (for merging layers, also when exporting a transparent PNGs)
-        # src (premultiplied) OVER dst (premultiplied)
-        # cB = cA + (1.0 - aA) * cB
-        #  where B = dst, A = src
-        srcAlpha = src[:,:,3:4].astype('float') * opacity / (1 << 15)
-        dstAlpha = dst[:,:,3:4].astype('float') / (1 << 15)
-        src_premult = src[:,:,0:3].astype('float') * opacity / (1<<15)
-        dst_premult = dst[:,:,0:3].astype('float') / (1<<15)
-        dst_c = clip(src_premult + (1.0 - srcAlpha) * dst_premult, 0.0, 1.0)
-        dst_a = clip(srcAlpha + dstAlpha - srcAlpha * dstAlpha, 0.0, 1.0)
-        dst[:,:,0:3] = clip(dst_c * (1<<15), 0, (1<<15) - 1).astype('uint16')
-        dst[:,:,3:4] = clip(dst_a * (1<<15), 0, (1<<15) - 1).astype('uint16')
-    elif dst.shape[2] == 3 and dst.dtype == 'uint16':
-        mypaintlib.tile_composite_rgba16_over_rgb16(src, dst, opacity)
-    else:
-        raise NotImplementedError
-
-
-def composite_array_multiply(dst, src, mipmap_level=0, opacity=1.0):
-    """The "svg:multiply" layer composite op implementation.
-    """
-    if dst.shape[2] == 4 and dst.dtype == 'uint16':
-        # rarely used (for merging layers, also when exporting a transparent PNGs)
-        # src (premultiplied) MULTIPLY dst (premultiplied)
-        # cA * cB +  cA * (1 - aB) + cB * (1 - aA)
-        srcAlpha = (opacity * src[:,:,3:4]).astype('float') / (1 << 15)
-        dstAlpha = (dst[:,:,3:4]).astype('float') / (1 << 15)
-        src_premult = src[:,:,0:3].astype('float') * opacity / (1<<15)
-        dst_premult = dst[:,:,0:3].astype('float') / (1<<15)
-        dst_c = clip(src_premult * dst_premult +  src_premult * (1.0 - dstAlpha) + dst_premult * (1.0 - srcAlpha),0.0, 1.0)
-        dst_a = clip(srcAlpha + dstAlpha - srcAlpha * dstAlpha, 0.0, 1.0)
-        dst[:,:,0:3] = clip(dst_c * (1<<15), 0, (1<<15) - 1).astype('uint16')
-        dst[:,:,3:4] = clip(dst_a * (1<<15), 0, (1<<15) - 1).astype('uint16')
-    elif dst.shape[2] == 3 and dst.dtype == 'uint16':
-        mypaintlib.tile_composite_rgba16_multiply_rgb16(src, dst, opacity)
-    else:
-        raise NotImplementedError
-
-
-def composite_array_screen(dst, src, mipmap_level=0, opacity=1.0):
-    """The "svg:screen" layer composite op implementation.
-    """
-    if dst.shape[2] == 4 and dst.dtype == 'uint16':
-        # rarely used (for merging layers, also when exporting a transparent PNGs)
-        # src (premultiplied) SCREEN dst (premultiplied)
-        # cA + cB - cA * cB
-        srcAlpha = (opacity * src[:,:,3:4]).astype('float') / (1 << 15)
-        dstAlpha = (dst[:,:,3:4]).astype('float') / (1 << 15)
-        src_premult = src[:,:,0:3].astype('float') * opacity / (1<<15)
-        dst_premult = dst[:,:,0:3].astype('float') / (1<<15)
-        dst_c = clip(src_premult + dst_premult - src_premult * dst_premult, 0, 1)
-        dst_a = clip(srcAlpha + dstAlpha - srcAlpha * dstAlpha, 0, 1)
-        dst[:,:,0:3] = clip(dst_c * (1<<15),0, (1<<15) - 1).astype('uint16')
-        dst[:,:,3:4] = clip(dst_a * (1<<15),0, (1<<15) - 1).astype('uint16')
-    elif dst.shape[2] == 3 and dst.dtype == 'uint16':
-        mypaintlib.tile_composite_rgba16_screen_rgb16(src, dst, opacity)
-    else:
-        raise NotImplementedError
-
-
-def composite_array_burn(dst, src, mipmap_level=0, opacity=1.0):
-    """The "svg:color-burn" layer composite op implementation.
-    """
-    if dst.shape[2] == 4 and dst.dtype == 'uint16':
-        # rarely used (for merging layers, also when exporting a transparent PNGs)
-        # src (premultiplied) OVER dst (premultiplied)
-        # if cA * aB + cB * aA <= aA * aB :
-        #   cA * (1 - aB) + cB * (1 - aA)
-        #   (cA == 0 ? 1 : (aA * (cA * aB + cB * aA - aA * aB) / cA) + cA * (1 - aB) + cB * (1 - aA))
-        #  where B = dst, A = src
-        aA = (opacity * src[:,:,3:4]).astype('float') / (1 << 15)
-        aB = (dst[:,:,3:4]).astype('float') / (1 << 15)
-        cA = src[:,:,0:3].astype('float') * opacity / (1<<15)
-        cB = dst[:,:,0:3].astype('float') / (1<<15)
-        dst_c = where(cA * aB + cB * aA <= aA * aB,
-                     cA * (1 - aB) + cB * (1 - aA),
-                     where(cA == 0,
-                           1.0,
-                           (aA * (cA * aB + cB * aA - aA * aB) / cA) + cA * (1.0 - aB) + cB * (1.0 - aA)))
-        dst_a = aA + aB - aA * aB
-        dst[:,:,0:3] = clip(dst_c * (1<<15), 0, (1<<15) - 1).astype('uint16')
-        dst[:,:,3:4] = clip(dst_a * (1<<15), 0, (1<<15) - 1).astype('uint16')
-    elif dst.shape[2] == 3 and dst.dtype == 'uint16':
-        mypaintlib.tile_composite_rgba16_burn_rgb16(src, dst, opacity)
-    else:
-        raise NotImplementedError
-
-
-def composite_array_dodge(dst, src, mipmap_level=0, opacity=1.0):
-    """The "svg:color-dodge" layer composite op implementation.
-    """
-    if dst.shape[2] == 4 and dst.dtype == 'uint16':
-        # rarely used (for merging layers, also when exporting a transparent PNGs)
-        # src (premultiplied) OVER dst (premultiplied)
-        # if cA * aB + cB * aA >= aA * aB :
-        #   aA * aB + cA * (1 - aB) + cB * (1 - aA)
-        #   (cA == aA ? 1 : cB * aA / (aA == 0 ? 1 : 1 - cA / aA)) + cA * (1 - aB) + cB * (1 - aA)
-        #  where B = dst, A = src
-        aA = (opacity * src[:,:,3:4]).astype('float') / (1 << 15)
-        aB = (dst[:,:,3:4]).astype('float') / (1 << 15)
-        cA = src[:,:,0:3].astype('float') * opacity / (1<<15)
-        cB = dst[:,:,0:3].astype('float') / (1<<15)
-        dst_c = where(cA * aB + cB * aA >= aA * aB,
-                     aA * aB + cA * (1 - aB) + cB * (1 - aA),
-                     where(cA == aA, 1.0, cB * aA / where(aA == 0, 1.0, 1.0 - cA / aA)) + cA * (1.0 - aB) + cB * (1.0 - aA))
-        dst_a = (aA + aB - aA * aB)
-        dst[:,:,0:3] = clip(dst_c * (1<<15),0, (1<<15) - 1).astype('uint16')
-        dst[:,:,3:4] = clip(dst_a * (1<<15),0, (1<<15) - 1).astype('uint16')
-    elif dst.shape[2] == 3 and dst.dtype == 'uint16':
-        mypaintlib.tile_composite_rgba16_dodge_rgb16(src, dst, opacity)
-    else:
-        raise NotImplementedError
-
+svg2composite_func = {
+    'svg:src-over': mypaintlib.tile_composite_normal,
+    'svg:multiply': mypaintlib.tile_composite_multiply,
+    'svg:screen': mypaintlib.tile_composite_screen,
+    'svg:overlay': mypaintlib.tile_composite_overlay,
+    'svg:darken': mypaintlib.tile_composite_darken,
+    'svg:lighten': mypaintlib.tile_composite_lighten,
+    'svg:hard-light': mypaintlib.tile_composite_hard_light,
+    'svg:soft-light': mypaintlib.tile_composite_soft_light,
+    'svg:color-burn': mypaintlib.tile_composite_color_burn,
+    'svg:color-dodge': mypaintlib.tile_composite_color_dodge,
+    'svg:difference': mypaintlib.tile_composite_difference,
+    'svg:exclusion': mypaintlib.tile_composite_exclusion,
+    'svg:hue': mypaintlib.tile_composite_hue,
+    'svg:saturation': mypaintlib.tile_composite_saturation,
+    'svg:color': mypaintlib.tile_composite_color,
+    'svg:luminosity': mypaintlib.tile_composite_luminosity,
+    }
 
 # tile for read-only operations on empty spots
 transparent_tile = Tile()
@@ -169,7 +76,62 @@ def get_tiles_bbox(tiles):
 class SurfaceSnapshot:
     pass
 
-class Surface(mypaintlib.TiledSurface):
+if use_gegl:
+
+    class GeglSurface(mypaintlib.GeglBackedSurface):
+
+        def __init__(self, mipmap_level=0):
+            mypaintlib.GeglBackedSurface.__init__(self, self)
+            self.observers = []
+
+        def notify_observers(self, *args):
+            for f in self.observers:
+                f(*args)
+
+        def get_bbox(self):
+            rect = helpers.Rect(*self.get_bbox_c())
+            return rect
+
+        def clear(self):
+            pass
+
+        def save_as_png(self, path, *args, **kwargs):
+            return self.save_as_png_c(str(path))
+
+        def load_from_png(self, path, x, y, *args, **kwargs):
+            return self.load_from_png_c(str(path))
+
+        def save_snapshot(self):
+            sshot = SurfaceSnapshot()
+            sshot.tiledict = {}
+            return sshot
+
+        def load_snapshot(self, sshot):
+            pass
+
+        def is_empty(self):
+            return False
+
+        def remove_empty_tiles(self):
+            pass
+
+        def composite_tile(self, dst, dst_has_alpha, tx, ty, mipmap_level=0, opacity=1.0,
+                           mode=DEFAULT_COMPOSITE_OP):
+            pass
+
+        def load_from_numpy(self, arr, x, y):
+            return (0, 0, 0, 0)
+
+        def load_from_surface(self, other):
+            pass
+
+        def get_tiles(self):
+            return {}
+
+        def set_symmetry_state(self, enabled, center_axis):
+            pass
+
+class MyPaintSurface(mypaintlib.TiledSurface):
     # the C++ half of this class is in tiledsurface.hpp
     def __init__(self, mipmap_level=0):
         mypaintlib.TiledSurface.__init__(self, self)
@@ -236,10 +198,11 @@ class Surface(mypaintlib.TiledSurface):
         if self.mipmap:
             self.mipmap._mark_mipmap_dirty(tx/2, ty/2)
 
-    def blit_tile_into(self, dst, tx, ty, mipmap_level=0):
+    def blit_tile_into(self, dst, dst_has_alpha, tx, ty, mipmap_level=0):
         # used mainly for saving (transparent PNG)
+        assert dst_has_alpha is True
         if self.mipmap_level < mipmap_level:
-            return self.mipmap.blit_tile_into(dst, tx, ty, mipmap_level)
+            return self.mipmap.blit_tile_into(dst, dst_has_alpha, tx, ty, mipmap_level)
         assert dst.shape[2] == 4
         src = self.get_tile_memory(tx, ty, readonly=True)
         if src is transparent_tile.rgba:
@@ -249,30 +212,20 @@ class Surface(mypaintlib.TiledSurface):
             mypaintlib.tile_convert_rgba16_to_rgba8(src, dst)
 
 
-    def composite_tile(self, dst, tx, ty, mipmap_level=0, opacity=1.0,
+    def composite_tile(self, dst, dst_has_alpha, tx, ty, mipmap_level=0, opacity=1.0,
                        mode=DEFAULT_COMPOSITE_OP):
         """Composite one tile of this surface over a NumPy array.
 
         Composite one tile of this surface over the array dst, modifying only dst.
         """
         if self.mipmap_level < mipmap_level:
-            return self.mipmap.composite_tile(dst, tx, ty, mipmap_level, opacity, mode)
+            return self.mipmap.composite_tile(dst, dst_has_alpha, tx, ty, mipmap_level, opacity, mode)
         if not (tx,ty) in self.tiledict:
             return
         src = self.get_tile_memory(tx, ty, readonly=True)
 
-        if mode == 'svg:src-over':
-            return composite_array_src_over(dst, src, mipmap_level, opacity)
-        elif mode == 'svg:multiply':
-            return composite_array_multiply(dst, src, mipmap_level, opacity)
-        elif mode == 'svg:screen':
-            return composite_array_screen(dst, src, mipmap_level, opacity)
-        elif mode == 'svg:color-burn':
-            return composite_array_burn(dst, src, mipmap_level, opacity)
-        elif mode == 'svg:color-dodge':
-            return composite_array_dodge(dst, src, mipmap_level, opacity)
-        else:
-            raise NotImplementedError, mode
+        func = svg2composite_func[mode]
+        func(src, dst, dst_has_alpha, opacity)
 
     def save_snapshot(self):
         sshot = SurfaceSnapshot()
@@ -308,17 +261,21 @@ class Surface(mypaintlib.TiledSurface):
 
         for tx, ty in s.get_tiles():
             dst = self.get_tile_memory(tx, ty, readonly=False)
-            s.blit_tile_into(dst, tx, ty)
+            s.blit_tile_into(dst, True, tx, ty)
 
         dirty_tiles.update(self.tiledict.keys())
         bbox = get_tiles_bbox(dirty_tiles)
         self.notify_observers(*bbox)
 
     def load_from_numpy(self, arr, x, y):
-        assert arr.dtype == 'uint8'
         h, w, channels = arr.shape
-        s = pixbufsurface.Surface(x, y, w, h, alpha=True, data=arr)
+        if h <= 0 or w <= 0:
+            return (x, y, w, h)
+
+        assert arr.dtype == 'uint8'
+        s = pixbufsurface.Surface(x, y, w, h, data=arr)
         self._load_from_pixbufsurface(s)
+        return (x, y, w, h)
 
     def load_from_png(self, filename, x, y, feedback_cb=None):
         """Load from a PNG, one tilerow at a time, discarding empty tiles.
@@ -329,8 +286,10 @@ class Surface(mypaintlib.TiledSurface):
         state = {}
         state['buf'] = None # array of height N, width depends on image
         state['ty'] = y/N # current tile row being filled into buf
+        state['frame_size'] = None
 
         def get_buffer(png_w, png_h):
+            state['frame_size'] = x, y, png_w, png_h
             if feedback_cb:
                 feedback_cb()
             buf_x0 = x/N*N
@@ -369,13 +328,17 @@ class Surface(mypaintlib.TiledSurface):
                     mypaintlib.tile_convert_rgba8_to_rgba16(src, dst)
 
         filename_sys = filename.encode(sys.getfilesystemencoding()) # FIXME: should not do that, should use open(unicode_object)
-        mypaintlib.load_png_fast_progressive(filename_sys, get_buffer)
+        flags = mypaintlib.load_png_fast_progressive(filename_sys, get_buffer)
         consume_buf() # also process the final chunk of data
-        
+        print flags
+
         dirty_tiles.update(self.tiledict.keys())
         bbox = get_tiles_bbox(dirty_tiles)
         self.notify_observers(*bbox)
 
+        # return the bbox of the loaded image
+        return state['frame_size']
+
     def render_as_pixbuf(self, *args, **kwargs):
         if not self.tiledict:
             print 'WARNING: empty surface'
@@ -404,3 +367,104 @@ class Surface(mypaintlib.TiledSurface):
         for pos, data in self.tiledict.items():
             if not data.rgba.any():
                 self.tiledict.pop(pos)
+
+    def get_move(self, x, y):
+        return _InteractiveMove(self, x, y)
+
+
+class _InteractiveMove:
+
+    def __init__(self, surface, x, y):
+        self.surface = surface
+        self.snapshot = surface.save_snapshot()
+        self.chunks = self.snapshot.tiledict.keys()
+        # print "Number of Tiledict_keys", len(self.chunks)
+        tx = x // N
+        ty = y // N
+        chebyshev = lambda p: max(abs(tx - p[0]), abs(ty - p[1]))
+        manhattan = lambda p: abs(tx - p[0]) + abs(ty - p[1])
+        euclidean = lambda p: math.sqrt((tx - p[0])**2 + (ty - p[1])**2)
+        self.chunks.sort(key=manhattan)
+        self.chunks_i = 0
+
+    def update(self, dx, dy):
+        # Tiles to be blanked at the end of processing
+        self.blanked = set(self.surface.tiledict.keys())
+        # Calculate offsets
+        self.slices_x = calc_translation_slices(int(dx))
+        self.slices_y = calc_translation_slices(int(dy))
+        self.chunks_i = 0
+
+    def cleanup(self):
+        # called at the end of each set of processing batches
+        for b in self.blanked:
+            self.surface.tiledict.pop(b, None)
+            self.surface._mark_mipmap_dirty(*b)
+        bbox = get_tiles_bbox(self.blanked)
+        self.surface.notify_observers(*bbox)
+        # Remove empty tile created by Layer Move
+        self.surface.remove_empty_tiles()
+
+    def process(self, n=200):
+        if self.chunks_i > len(self.chunks):
+            return False
+        written = set()
+        if n <= 0:
+            n = len(self.chunks)  # process all remaining
+        for tile_pos in self.chunks[self.chunks_i : self.chunks_i + n]:
+            src_tx, src_ty = tile_pos
+            src_tile = self.snapshot.tiledict[(src_tx, src_ty)]
+            is_integral = len(self.slices_x) == 1 and len(self.slices_y) == 1
+            for (src_x0, src_x1), (targ_tdx, targ_x0, targ_x1) in self.slices_x:
+                for (src_y0, src_y1), (targ_tdy, targ_y0, targ_y1) in self.slices_y:
+                    targ_tx = src_tx + targ_tdx
+                    targ_ty = src_ty + targ_tdy
+                    if is_integral:
+                        self.surface.tiledict[(targ_tx, targ_ty)] = src_tile.copy()
+                    else:
+                        targ_tile = None
+                        if (targ_tx, targ_ty) in self.blanked:
+                            targ_tile = Tile()
+                            self.surface.tiledict[(targ_tx, targ_ty)] = targ_tile
+                            self.blanked.remove( (targ_tx, targ_ty) )
+                        else:
+                            targ_tile = self.surface.tiledict.get((targ_tx, targ_ty), None)
+                        if targ_tile is None:
+                            targ_tile = Tile()
+                            self.surface.tiledict[(targ_tx, targ_ty)] = targ_tile
+                        targ_tile.rgba[targ_y0:targ_y1, targ_x0:targ_x1] \
+                          = src_tile.rgba[src_y0:src_y1, src_x0:src_x1]
+                    written.add((targ_tx, targ_ty))
+        self.blanked -= written
+        for pos in written:
+            self.surface._mark_mipmap_dirty(*pos)
+        bbox = get_tiles_bbox(written) # hopefully relatively contiguous
+        self.surface.notify_observers(*bbox)
+        self.chunks_i += n
+        return self.chunks_i < len(self.chunks)
+
+
+def calc_translation_slices(dc):
+    """Returns a list of offsets and slice extents for a translation of `dc`.
+
+    The returned slice list's members are of the form
+
+        ((src_c0, src_c1), (targ_tdc, targ_c0, targ_c1))
+
+    where ``src_c0`` and ``src_c1`` determine the extents of the source slice
+    within a tile, their ``targ_`` equivalents specify where to put that slice
+    in the target tile, and ``targ_tdc`` is the tile offset.
+    """
+    dcr = dc % N
+    tdc = (dc // N)
+    if dcr == 0:
+        return [ ((0, N), (tdc, 0, N)) ]
+    else:
+        return [ ((0, N-dcr), (tdc, dcr, N)) ,
+                 ((N-dcr, N), (tdc+1, 0, dcr)) ]
+
+# Set which surface backend to use
+Surface = GeglSurface if use_gegl else MyPaintSurface
+
+def new_surface():
+    return Surface()
