@@ -74,17 +74,16 @@ def paint(gui):
 
     b = gui.app.brushmanager.get_brush_by_name('redbrush')
     assert b, 'brush not found'
-    gui.app.brushmanager.select_brush(b)
 
     dw.fullscreen_cb()
     gui.wait_for_idle()
+    gui.app.brushmanager.select_brush(b)
     gui.wait_for_duration(1.5) # fullscreen seems to take some time to get through...
     gui.wait_for_idle()
 
     events = loadtxt('painting30sec.dat.gz')
     events = list(events)
     yield start_measurement
-    t0 = time()
     t_old = 0.0
     t_last_redraw = 0.0
     for t, x, y, pressure in events:
@@ -93,8 +92,7 @@ def paint(gui):
             t_last_redraw = t
         dtime = t - t_old
         t_old = t
-        cr = tdw.get_model_coordinates_cairo_context()
-        x, y = cr.device_to_user(x, y)
+        x, y = tdw.display_to_model(x, y)
         gui_doc.model.stroke_to(dtime, x, y, pressure, 0.0, 0.0)
     yield stop_measurement
 
@@ -177,9 +175,37 @@ def save_png_layer():
     d = document.Document()
     d.load('biglayer.png')
     yield start_measurement
-    d.layer.surface.save('test_save.png')
+    d.layer.save_as_png('test_save.png')
     yield stop_measurement
 
+
+@nogui_test
+def brushengine_paint_hires():
+    from lib import tiledsurface, brush
+    s = tiledsurface.Surface()
+    bi = brush.BrushInfo(open('brushes/watercolor.myb').read())
+    b = brush.Brush(bi)
+
+    events = loadtxt('painting30sec.dat.gz')
+    t_old = events[0][0]
+    yield start_measurement
+    s.begin_atomic()
+    trans_time = 0.0
+    for t, x, y, pressure in events:
+        dtime = t - t_old
+        t_old = t
+        b.stroke_to (s, x*5, y*5, pressure, 0.0, 0.0, dtime)
+
+        trans_time += dtime
+        if trans_time > 0.05:
+            trans_time = 0.0
+            s.end_atomic()
+            s.begin_atomic()
+
+    s.end_atomic()
+    yield stop_measurement
+    #s.save('test_paint_hires.png') # approx. 3000x3000
+
 @gui_test
 def scroll_nozoom(gui):
     gui.wait_for_idle()
@@ -324,7 +350,7 @@ if __name__ == '__main__':
                     fname = '%s_%s_%d.pstats' % (options.profile, t, i)
                 args[3] = fname
             child = subprocess.Popen(args, stdout=subprocess.PIPE)
-            output, trash = child.communicate()
+            output, junk = child.communicate()
             if child.returncode != 0:
                 print 'FAILED'
                 break
